import asyncio
import uuid

import asynctnt
from asynctnt._testbase import check_version
from asynctnt.connection import ConnectionState
from asynctnt.exceptions import TarantoolDatabaseError, ErrorCode, \
    TarantoolNotConnectedError
from asynctnt.instance import TarantoolSyncInstance
from tests import BaseTarantoolTestCase


class ConnectTestCase(BaseTarantoolTestCase):
    DO_CONNECT = False

    async def test__connect(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0)
        self.assertEqual(conn.host, self.tnt.host)
        self.assertEqual(conn.port, self.tnt.port)
        self.assertIsNone(conn.username)
        self.assertIsNone(conn.password)
        self.assertEqual(conn.reconnect_timeout, 0)
        self.assertEqual(conn.connect_timeout, 3)
        self.assertEqual(conn.loop, self.loop)
        self.assertIsNone(conn.initial_read_buffer_size)
        self.assertIsNone(conn.schema_id)
        self.assertIsNone(conn.version)
        self.assertEqual(
            repr(conn),
            "<asynctnt.Connection host={} port={} state={}>".format(
                conn.host,
                conn.port,
                repr(conn.state)
            ))
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

        c = await conn.connect()
        self.assertEqual(c, conn)
        self.assertIsNotNone(conn._transport)
        self.assertIsNotNone(conn._protocol)
        self.assertTrue(conn.is_connected)
        self.assertTrue(conn.is_fully_connected)
        self.assertEqual(conn.state, ConnectionState.CONNECTED)
        self.assertIsNotNone(conn._protocol.schema)
        self.assertIsNotNone(conn.version)

        await conn.call('box.info')
        await conn.disconnect()

    async def test__connect_direct(self):
        conn = await asynctnt.connect(host=self.tnt.host, port=self.tnt.port,
                                      reconnect_timeout=0)
        self.assertEqual(conn.host, self.tnt.host)
        self.assertEqual(conn.port, self.tnt.port)
        self.assertIsNone(conn.username)
        self.assertIsNone(conn.password)
        self.assertEqual(conn.reconnect_timeout, 0)
        self.assertEqual(conn.connect_timeout, 3)
        self.assertEqual(conn.loop, self.loop)
        self.assertIsNone(conn.initial_read_buffer_size)

        self.assertIsNotNone(conn._transport)
        self.assertIsNotNone(conn._protocol)
        self.assertTrue(conn.is_connected)
        self.assertTrue(conn.is_fully_connected)
        self.assertEqual(conn.state, ConnectionState.CONNECTED)
        self.assertIsNotNone(conn._protocol.schema)
        self.assertIsNotNone(conn.version)

        await conn.call('box.info')
        await conn.disconnect()

    async def test__connect_unix(self):
        if self.in_docker:
            self.skipTest('Skipping as running inside the docker')
            return

        tnt = TarantoolSyncInstance(
            host='unix/',
            port='/tmp/' + uuid.uuid4().hex + '.sock',
            console_host='127.0.0.1',
            applua=self.read_applua(),
            cleanup=self.TNT_CLEANUP
        )
        tnt.start()
        try:
            conn = await asynctnt.connect(host=tnt.host, port=tnt.port,
                                          reconnect_timeout=0)
            self.assertEqual(conn.host, tnt.host)
            self.assertEqual(conn.port, tnt.port)
            self.assertIsNone(conn.username)
            self.assertIsNone(conn.password)
            self.assertEqual(conn.reconnect_timeout, 0)
            self.assertEqual(conn.connect_timeout, 3)
            self.assertEqual(conn.loop, self.loop)
            self.assertIsNone(conn.initial_read_buffer_size)

            self.assertIsNotNone(conn._transport)
            self.assertIsNotNone(conn._protocol)
            self.assertTrue(conn.is_connected)
            self.assertTrue(conn.is_fully_connected)
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertIsNotNone(conn._protocol.schema)
            self.assertIsNotNone(conn.version)

            await conn.call('box.info')
            await conn.disconnect()
        finally:
            tnt.stop()

    async def test__connect_contextmanager(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0)
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

        async with conn:
            self.assertIsNotNone(conn._transport)
            self.assertIsNotNone(conn._protocol)
            self.assertTrue(conn.is_connected)
            self.assertTrue(conn.is_fully_connected)
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertIsNotNone(conn._protocol.schema)
            self.assertIsNotNone(conn.version)

            await conn.call('box.info')

        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

    async def test__connect_contextmanager_connect_inside(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0)
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

        async with conn:
            await conn.connect()
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            await conn.call('box.info')

        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

    async def test__connect_contextmanager_disconnect_inside(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0)
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

        async with conn:
            await conn.disconnect()
            self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

            with self.assertRaises(TarantoolNotConnectedError):
                await conn.call('box.info')

        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

    async def test__connect_no_schema(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0,
                                   fetch_schema=False,
                                   auto_refetch_schema=False)
        async with conn:
            self.assertIsNotNone(conn._transport)
            self.assertIsNotNone(conn._protocol)
            self.assertTrue(conn.is_connected)
            self.assertTrue(conn.is_fully_connected)
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertIsNotNone(conn._protocol.schema)
            await conn.call('box.info')

    async def test__connect_auth(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   username='t1', password='t1',
                                   reconnect_timeout=0)
        async with conn:
            self.assertIsNotNone(conn._transport)
            self.assertIsNotNone(conn._protocol)
            self.assertTrue(conn.is_connected)
            self.assertTrue(conn.is_fully_connected)
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertIsNotNone(conn._protocol.schema)
            await conn.call('box.info')

    async def test__connect_auth_no_schema(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   username='t1', password='t1',
                                   fetch_schema=False,
                                   auto_refetch_schema=False,
                                   reconnect_timeout=0)
        self.assertEqual(conn.username, 't1')
        self.assertEqual(conn.password, 't1')
        async with conn:
            self.assertIsNotNone(conn._transport)
            self.assertIsNotNone(conn._protocol)
            self.assertTrue(conn.is_connected)
            self.assertTrue(conn.is_fully_connected)
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertIsNotNone(conn._protocol.schema)
            await conn.call('box.info')

    async def test__disconnect(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0)
        await conn.connect()
        await conn.disconnect()
        self.assertFalse(conn.is_connected)
        self.assertFalse(conn.is_fully_connected)
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

        with self.assertRaises(TarantoolNotConnectedError):
            await conn.call('box.info')

    async def test__disconnect_in_request(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0)
        await conn.connect()

        coro = self.ensure_future(conn.eval('require "fiber".sleep(2)'))
        await self.sleep(0.5)
        await conn.disconnect()

        with self.assertRaises(TarantoolNotConnectedError):
            await coro

    async def test__disconnect_auth(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   username='t1', password='t1',
                                   reconnect_timeout=0)
        await conn.connect()
        await conn.disconnect()
        self.assertFalse(conn.is_connected)
        self.assertFalse(conn.is_fully_connected)
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

        with self.assertRaises(TarantoolNotConnectedError):
            await conn.call('box.info')

    async def test__disconnect_while_reconnecting(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   username='t1', password='t1',
                                   reconnect_timeout=0.1)
        self.assertEqual(conn.reconnect_timeout, 0.1)
        try:
            await conn.connect()
            self.tnt.stop()
            await self.sleep(0.5)

            await conn.disconnect()

            self.assertFalse(conn.is_connected)
            self.assertFalse(conn.is_fully_connected)
            self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

            with self.assertRaises(TarantoolNotConnectedError):
                await conn.call('box.info')
        finally:
            self.tnt.start()

    async def test__close_while_reconnecting(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   username='t1', password='t1',
                                   reconnect_timeout=0.1)
        self.assertEqual(conn.reconnect_timeout, 0.1)
        try:
            await conn.connect()
            self.tnt.stop()
            await self.sleep(0.5)

            conn.close()

            self.assertFalse(conn.is_connected)
            self.assertFalse(conn.is_fully_connected)
            self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

            with self.assertRaises(TarantoolNotConnectedError):
                await conn.call('box.info')
        finally:
            self.tnt.start()

    async def test__connect_multiple(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   fetch_schema=False,
                                   reconnect_timeout=0)
        for _ in range(10):
            await conn.connect()
            await conn.disconnect()
        self.assertFalse(conn.is_connected)
        self.assertFalse(conn.is_fully_connected)
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

        with self.assertRaises(TarantoolNotConnectedError):
            await conn.call('box.info')

    async def test__connect_cancel(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   fetch_schema=True,
                                   reconnect_timeout=0)
        try:
            f = asyncio.ensure_future(conn.connect())
            await self.sleep(0.0001)
            f.cancel()
            with self.assertRaises(asyncio.CancelledError):
                await f
        finally:
            await conn.disconnect()

    async def test__connect_error_no_reconnect(self):
        conn = asynctnt.Connection(host="127.0.0.1", port=1,
                                   fetch_schema=True,
                                   reconnect_timeout=0)
        with self.assertRaises(ConnectionRefusedError):
            await conn.connect()

    async def test__connect_wait_tnt_started(self):
        self.tnt.stop()
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   username='t1', password='t1',
                                   fetch_schema=True,
                                   reconnect_timeout=0.000000001)
        try:
            coro = self.ensure_future(conn.connect())
            await self.sleep(0.3)
            self.tnt.start()
            await self.sleep(1)
            while True:
                try:
                    await coro
                    break
                except TarantoolDatabaseError as e:
                    if e.code == ErrorCode.ER_NO_SUCH_USER:
                        # Try again
                        coro = self.ensure_future(conn.connect())
                        continue
                    raise

            self.assertEqual(conn.state, ConnectionState.CONNECTED)

            await conn.call('box.info')
        finally:
            await conn.disconnect()

    async def test__connect_waiting_for_spaces(self):
        if self.in_docker:
            self.skipTest('not running in docker')
            return

        with self.make_instance() as tnt:
            tnt.replication_source = ['x:1']
            tnt.start(wait=False)

            conn = asynctnt.Connection(host=tnt.host, port=tnt.port,
                                       fetch_schema=True,
                                       reconnect_timeout=0.1,
                                       connect_timeout=10)
            self.assertEqual(conn.connect_timeout, 10)
            try:
                states = {}

                async def state_checker():
                    while True:
                        states[conn.state] = True
                        await self.sleep(0.001)

                checker = self.ensure_future(state_checker())

                try:
                    await asyncio.wait_for(conn.connect(), 1)
                except asyncio.TimeoutError:
                    self.assertTrue(True, 'connect cancelled')

                checker.cancel()

                self.assertTrue(states.get(ConnectionState.CONNECTING, False),
                                'was in connecting')

                with self.assertRaises(TarantoolNotConnectedError):
                    await conn.call('box.info')
            finally:
                await conn.disconnect()

    async def test__connect_waiting_for_spaces_no_reconnect(self):
        if self.in_docker:
            self.skipTest('not running in docker')
            return

        with self.make_instance() as tnt:
            tnt.replication_source = ['x:1']
            tnt.start(wait=False)
            await self.sleep(1)

            if not check_version(self, tnt.version(), min=(1, 7)):
                return

            conn = asynctnt.Connection(host=tnt.host, port=tnt.port,
                                       fetch_schema=True,
                                       reconnect_timeout=0,
                                       connect_timeout=10)
            try:
                with self.assertRaises(TarantoolDatabaseError) as e:
                    await conn.connect()

                self.assertEqual(e.exception.code, ErrorCode.ER_NO_SUCH_SPACE)
            finally:
                await conn.disconnect()

    async def test__connect_waiting_for_spaces_no_reconnect_1_6(self):
        with self.make_instance() as tnt:
            tnt.replication_source = ['x:1']
            tnt.start(wait=False)
            await self.sleep(1)

            if not check_version(self, tnt.version(), max=(1, 7)):
                return

            conn = asynctnt.Connection(host=tnt.host, port=tnt.port,
                                       fetch_schema=True,
                                       reconnect_timeout=0,
                                       connect_timeout=10)
            try:
                with self.assertRaises(ConnectionRefusedError):
                    await conn.connect()
            finally:
                await conn.disconnect()

    async def test__connect_err_loading(self):
        if self.in_docker:
            self.skipTest('not running in docker')
            return

        with self.make_instance() as tnt:
            tnt.replication_source = ['x:1']
            tnt.start(wait=False)
            await self.sleep(1)

            if not check_version(self, tnt.version(), min=(1, 7)):
                return

            conn = asynctnt.Connection(host=tnt.host, port=tnt.port,
                                       username='t1', password='t1',
                                       fetch_schema=True,
                                       reconnect_timeout=0,
                                       connect_timeout=10)
            try:
                with self.assertRaises(TarantoolDatabaseError) as e:
                    await conn.connect()

                self.assertEqual(e.exception.code, ErrorCode.ER_LOADING)
            finally:
                await conn.disconnect()

    async def test__connect_err_loading_1_6(self):
        with self.make_instance() as tnt:
            tnt.replication_source = ['x:1']
            tnt.start(wait=False)

            await self.sleep(1)
            if not check_version(self, tnt.version(), max=(1, 7)):
                return

            conn = asynctnt.Connection(host=tnt.host, port=tnt.port,
                                       username='t1', password='t1',
                                       fetch_schema=True,
                                       reconnect_timeout=0,
                                       connect_timeout=10)
            try:
                with self.assertRaises(ConnectionRefusedError):
                    await conn.connect()
            finally:
                await conn.disconnect()

    async def test__connect_tnt_restarted(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   username='t1', password='t1',
                                   fetch_schema=True,
                                   reconnect_timeout=0.000001)
        await conn.connect()

        try:
            self.tnt.stop()
            self.tnt.start()
            await self.sleep(0.5)
            await conn.ping()
        except Exception as e:
            self.fail(
                'Should not throw any exceptions, but got: {}'.format(e))
        finally:
            await conn.disconnect()

    async def test__connect_force_disconnect(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=44444,
                                   reconnect_timeout=0.3)
        self.ensure_future(conn.connect())
        await self.sleep(1)
        await conn.disconnect()
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

    async def test__close(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0)
        await conn.connect()
        await self.sleep(0.1)
        conn.close()
        await self.sleep(0.1)
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

    async def test_disconnect_from_idle(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0)
        await conn.disconnect()
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

    async def test_reconnect_from_idle(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0)
        await conn.reconnect()
        try:

            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            await conn.call('box.info')
        finally:
            await conn.disconnect()

    async def test_reconnect_after_connect(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0)
        try:
            await conn.connect()
            await conn.reconnect()

            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            await conn.call('box.info')
        finally:
            await conn.disconnect()

    async def test_manual_reconnect(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=0)
        try:
            await conn.connect()
            await conn.disconnect()
            await conn.connect()

            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            await conn.call('box.info')
        finally:
            await conn.disconnect()

    async def test__connect_connection_lost(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=1/3)
        try:
            await conn.connect()
            self.tnt.stop()
            await self.sleep(0.5)
            self.tnt.start()
            await self.sleep(0.5)

            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertTrue(conn.is_connected)
            await conn.call('box.info')
        finally:
            await conn.disconnect()

    async def test__connect_from_multiple_coroutines(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=1/3)
        try:
            coros = []
            for _ in range(10):
                coros.append(asyncio.ensure_future(conn.connect()))

            await asyncio.gather(*coros)
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertTrue(conn.is_connected)
            await conn.call('box.info')
        finally:
            await conn.disconnect()

    async def test__disconnect_from_multiple_coroutines(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=1/3)
        try:
            await conn.connect()
            coros = []
            for _ in range(10):
                coros.append(asyncio.ensure_future(conn.disconnect()))

            await asyncio.gather(*coros)
            self.assertEqual(conn.state, ConnectionState.DISCONNECTED)
            self.assertFalse(conn.is_connected)

            with self.assertRaises(TarantoolNotConnectedError):
                await conn.call('box.info')
        finally:
            await conn.disconnect()

    async def test__connect_while_reconnecting(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   reconnect_timeout=1)

        try:
            coros = []
            for _ in range(10):
                coros.append(asyncio.ensure_future(conn.connect()))

            self.tnt.stop()
            await self.sleep(0.5)

            connect_coros = asyncio.ensure_future(asyncio.gather(*coros))

            self.tnt.start()
            await self.sleep(1)
            await connect_coros

            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertTrue(conn.is_connected)

            await conn.call('box.info')
        finally:
            await conn.disconnect()

    async def test__connect_on_tnt_crash_no_reconnect(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   connect_timeout=1,
                                   reconnect_timeout=0)
        try:
            await conn.connect()
            await asyncio.sleep(0.5)
            try:
                await conn.eval("require('ffi').cast('char *', 0)[0] = 48")
            except TarantoolNotConnectedError:
                self.assertTrue(True, 'not connected error triggered')
            self.tnt.stop()
            self.tnt.start()
            await asyncio.sleep(1)
            await conn.connect()  # this connect should reconnect easily

            await conn.call('box.info')
        finally:
            await conn.disconnect()

    async def test__connect_on_tnt_crash_with_reconnect(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   connect_timeout=1,
                                   reconnect_timeout=1/3)
        try:
            await conn.connect()
            await asyncio.sleep(0.5)
            try:
                await conn.eval("require('ffi').cast('char *', 0)[0] = 48")
            except TarantoolNotConnectedError:
                self.assertTrue(True, 'not connected error triggered')
            self.tnt.stop()
            self.tnt.start()
            await asyncio.sleep(1)
            await conn.connect()  # this connect should reconnect easily

            await conn.call('box.info')
        finally:
            await conn.disconnect()

    async def test__connect_invalid_user_no_reconnect(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   username='fancy', password='man',
                                   connect_timeout=1,
                                   reconnect_timeout=0)
        with self.assertRaises(TarantoolDatabaseError) as e:
            await conn.connect()

        self.assertEqual(e.exception.code, ErrorCode.ER_NO_SUCH_USER)

    async def test__connect_invalid_user_with_reconnect(self):
        conn = asynctnt.Connection(host=self.tnt.host, port=self.tnt.port,
                                   fetch_schema=True,
                                   reconnect_timeout=0.1,
                                   connect_timeout=10)
        await conn.connect()  # first connect successfully

        # then change credentials
        conn._username = 'fancy'
        conn._password = 'man'

        self.tnt.stop()
        self.tnt.start()
        await self.sleep(0.1)
        try:
            states = {}

            async def state_checker():
                while True:
                    states[conn.state] = True
                    await self.sleep(0.001)

            checker = self.ensure_future(state_checker())

            try:
                await asyncio.wait_for(conn.connect(), 1)
            except asyncio.TimeoutError:
                self.assertTrue(True, 'connect cancelled')

            checker.cancel()

            self.assertTrue(states.get(ConnectionState.CONNECTING, False),
                            'was in connecting')
            self.assertTrue(states.get(ConnectionState.RECONNECTING, False),
                            'was in connecting')

            with self.assertRaises(TarantoolNotConnectedError):
                await conn.call('box.info')
        finally:
            await conn.disconnect()



from asynctnt import Response
from asynctnt._testbase import ensure_version
from tests import BaseTarantoolTestCase


class SQLTestCase(BaseTarantoolTestCase):
    @ensure_version(min=(2, 0))
    async def test__sql_basic(self):
        res = await self.conn.sql('select 1, 2')

        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, [[1, 2]], 'Body ok')

    @ensure_version(min=(2, 0))
    async def test__sql_with_param(self):
        res = await self.conn.sql('select 1, 2 where 1 = ?', [1])

        self.assertResponseEqual(res, [[1, 2]], 'Body ok')

    @ensure_version(min=(2, 0))
    async def test__sql_with_param_cols(self):
        res = await self.conn.sql('select 1 as a, 2 as b where 1 = ?', [1])

        self.assertResponseEqualKV(res, [{'A': 1, 'B': 2}], 'Body ok')

    @ensure_version(min=(2, 0))
    async def test__sql_insert(self):
        res = await self.conn.sql(
            "insert into sql_space (id, name) values (1, 'one')")
        self.assertEqual(1, res.rowcount, 'rowcount ok')

    @ensure_version(min=(2, 0))
    async def test__sql_empty_autoincrement(self):
        res = await self.conn.sql(
            "insert into sql_space (id, name) values (1, 'one')")
        self.assertEqual(None, res.autoincrement_ids, 'autoincrement ok')

    @ensure_version(min=(2, 0))
    async def test__sql_insert_autoincrement(self):
        res = await self.conn.sql(
            "insert into sql_space_autoincrement (name) values ('name')")
        self.assertEqual(1, res.rowcount, 'rowcount ok')
        self.assertEqual([1], res.autoincrement_ids, 'autoincrement ok')

    @ensure_version(min=(2, 0))
    async def test__sql_insert_autoincrement_multiple(self):
        res = await self.conn.sql(
            "insert into sql_space_autoincrement_multiple (name) values ('name'), ('name2')")
        self.assertEqual(2, res.rowcount, 'rowcount ok')
        self.assertEqual([1, 2], res.autoincrement_ids, 'autoincrement ok')

    @ensure_version(min=(2, 0))
    async def test__sql_insert_multiple(self):
        res = await self.conn.sql(
            "insert into sql_space (id, name) "
            "values (1, 'one'), (2, 'two')")
        self.assertEqual(2, res.rowcount, 'rowcount ok')

    @ensure_version(min=(2, 0))
    async def test__sql_update(self):
        await self.conn.sql("insert into sql_space values (1, 'one')")

        res = await self.conn.sql(
            "update sql_space set name = 'uno' where id = 1")

        self.assertEqual(1, res.rowcount, 'rowcount ok')

    @ensure_version(min=(2, 0))
    async def test__sql_update_multiple(self):
        await self.conn.sql("insert into sql_space values (1, 'one')")
        await self.conn.sql("insert into sql_space values (2, 'two')")

        res = await self.conn.sql("update sql_space set name = 'uno'")

        self.assertEqual(2, res.rowcount, 'rowcount ok')

    @ensure_version(min=(2, 0))
    async def test__sql_delete(self):
        await self.conn.sql("insert into sql_space values (1, 'one')")
        res = await self.conn.sql("delete from sql_space where name = 'one'")
        self.assertEqual(1, res.rowcount, 'rowcount ok')

    @ensure_version(min=(2, 0))
    async def test__sql_delete_multiple(self):
        await self.conn.sql("insert into sql_space values (1, 'one')")
        await self.conn.sql("insert into sql_space values (2, 'two')")

        res = await self.conn.sql("delete from sql_space")
        # TODO: wait for https://github.com/tarantool/tarantool/issues/3816
        # self.assertEqual(2, res.rowcount, 'rowcount ok')

        res = await self.conn.sql("select * from sql_space")
        self.assertEqual(0, res.rowcount, 'rowcount is surely ok')



cimport asynctnt.iproto.tupleobj as tupleobj
cimport asynctnt.iproto.tarantool as tarantool

include "const.pxi"

include "cmsgpuck.pxd"
include "python.pxd"

include "unicodeutil.pxd"
include "schema.pxd"
include "buffer.pxd"
include "rbuffer.pxd"
include "request.pxd"
include "response.pxd"
include "db.pxd"
include "push.pxd"

include "coreproto.pxd"


cdef class BaseProtocol(CoreProtocol):
    cdef:
        object loop
        str username
        str password
        bint fetch_schema
        bint auto_refetch_schema
        float request_timeout

        object connected_fut
        object on_connection_made_cb
        object on_connection_lost_cb

        object _on_request_completed_cb
        object _on_request_timeout_cb

        dict _reqs
        uint64_t _sync
        Schema _schema
        int64_t _schema_id
        bint _schema_fetch_in_progress
        object _refetch_schema_future
        Db _db

        object create_future

    cdef void _set_connection_ready(self)
    cdef void _set_connection_error(self, e)

    cdef void _do_auth(self, str username, str password)
    cdef void _do_fetch_schema(self, object fut)
    cdef object _refetch_schema(self)

    cdef inline uint64_t next_sync(self)
    cdef uint32_t transform_iterator(self, iterator) except *

    cdef object _new_waiter_for_request(self, Request req, float timeout)
    cdef Db _create_db(self)
    cdef object execute(self, Request req, WriteBuffer buf, float timeout)



import logging
import os
import sys

from asynctnt._testbase import TarantoolTestCase

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


class BaseTarantoolTestCase(TarantoolTestCase):
    DO_CONNECT = True
    LOGGING_LEVEL = getattr(logging, os.getenv('LOG', 'CRITICAL').upper())
    LOGGING_STREAM = sys.stdout
    TNT_APP_LUA_PATH = os.path.join(CURRENT_DIR, 'files', 'app.lua')

    TESTER_SPACE_ID = 512
    TESTER_SPACE_NAME = 'tester'

    async def truncate(self):
        if self.conn and self.conn.is_connected:
            await self.conn.call('truncate', timeout=5)

    def tearDown(self):
        if hasattr(self, 'conn'):
            self.loop.run_until_complete(self.truncate())
        super().tearDown()



from libc.stdint cimport uint64_t, int64_t


cdef public class TntFields [object C_TntFields, type C_TntFields_Type]:
    cdef:
        list _names  # contains only field names ('f1', 'f2', ...)
        dict _mapping  # contains field's name => id mappings

    cdef inline int len(self)
    cdef inline void add(self, uint64_t id, str name)
    cdef inline str name_by_id(self, int i)
    cdef inline uint64_t id_by_name(self, str f) except *


cdef class SchemaIndex:
    cdef:
        int sid
        int iid
        str name
        str index_type
        object unique

        TntFields fields


cdef class SchemaDummyIndex(SchemaIndex):
    pass


cdef class SchemaSpace:
    cdef:
        int sid
        int owner
        str name
        str engine
        int field_count
        object flags

        TntFields fields
        dict indexes

    cdef void add_index(self, SchemaIndex idx)
    cdef SchemaIndex get_index(self, index, create_dummy=*)


cdef class SchemaDummySpace(SchemaSpace):
    pass


cdef class Schema:
    cdef:
        dict schema
        int64_t id

    cdef SchemaSpace get_space(self, space)
    cdef SchemaSpace create_dummy_space(self, int space_id)
    cdef SchemaSpace get_or_create_space(self, space)

    cdef SchemaSpace parse_space(self, space_row)
    cdef SchemaIndex parse_index(self, index_row)

    cdef inline clear(self)

    @staticmethod
    cdef Schema parse(int64_t schema_id, spaces, indexes)


cdef list dict_to_list_fields(dict d, TntFields fields, bint default_none)



cimport cpython
cimport cython
cimport cpython.bytes
cimport cpython.list
cimport cpython.tuple
cimport cpython.dict
cimport cpython.unicode

from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
from cpython.ref cimport PyObject

from libc.string cimport memcpy
from libc.stdint cimport uint32_t, uint64_t, int64_t


# noinspection PyUnresolvedReferences
# noinspection PyAttributeOutsideInit
@cython.no_gc_clear
@cython.final
@cython.freelist(_BUFFER_FREELIST_SIZE)
cdef class WriteBuffer:
    def __cinit__(self):
        self._smallbuf_inuse = True
        self._buf = self._smallbuf
        self._size = _BUFFER_INITIAL_SIZE
        self._length = 0
        self._encoding = None
        self.__op_offset = -1
        self.__sync_offset = -1
        self.__schema_id_offset = -1

    def __dealloc__(self):
        if self._buf is not NULL and not self._smallbuf_inuse:
            PyMem_Free(self._buf)
            self._buf = NULL
            self._size = 0

        if self._view_count:
            raise RuntimeError(
                'Deallocating buffer with attached memoryviews')

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        self._view_count += 1

        cpython.PyBuffer_FillInfo(
            buffer, self, self._buf, self._length,
            1,  # read-only
            flags
        )

    def __releasebuffer__(self, Py_buffer *buffer):
        self._view_count -= 1

    def hex(self):
        return ":".join("{:02x}".format(ord(<bytes>(self._buf[i])))
                        for i in range(self._length))

    @staticmethod
    cdef WriteBuffer new(bytes encoding):
        cdef WriteBuffer buf
        buf = WriteBuffer.__new__(WriteBuffer)
        buf._encoding = encoding
        return buf

    cdef inline _check_readonly(self):
        if self._view_count:
            raise BufferError('the buffer is in read-only mode')

    cdef inline len(self):
        return self._length

    cdef void ensure_allocated(self, ssize_t extra_length) except *:
        cdef ssize_t new_size = extra_length + self._length

        if new_size > self._size:
            self._reallocate(new_size)

    cdef char *_ensure_allocated(self, char *p,
                                 ssize_t extra_length) except NULL:
        cdef:
            ssize_t new_size

        new_size = extra_length + self._length

        if new_size > self._size:
            used = p - self._buf
            self._reallocate(new_size)
            p = &self._buf[used]
        return p

    cdef void _reallocate(self, ssize_t new_size) except *:
        cdef char *new_buf

        if new_size < _BUFFER_MAX_GROW:
            new_size = _BUFFER_MAX_GROW
        else:
            # Add a little extra
            new_size += _BUFFER_INITIAL_SIZE

        if self._smallbuf_inuse:
            new_buf = <char*>PyMem_Malloc(sizeof(char) * <size_t>new_size)
            if new_buf is NULL:
                self._buf = NULL
                self._size = 0
                self._length = 0
                raise MemoryError
            memcpy(new_buf, self._buf, <size_t>self._size)
            self._size = new_size
            self._buf = new_buf
            self._smallbuf_inuse = False
        else:
            new_buf = <char*>PyMem_Realloc(<void*>self._buf, <size_t>new_size)
            if new_buf is NULL:
                PyMem_Free(self._buf)
                self._buf = NULL
                self._size = 0
                self._length = 0
                raise MemoryError
            self._buf = new_buf
            self._size = new_size

    cdef void write_buffer(self, WriteBuffer buf) except *:
        if not buf._length:
            return

        self.ensure_allocated(buf._length)
        memcpy(self._buf + self._length,
               <void*>buf._buf,
               <size_t>buf._length)
        self._length += buf._length

    cdef void write_header(self, uint64_t sync,
                           tarantool.iproto_type op,
                           int64_t schema_id) except *:
        cdef:
            char *begin = NULL
            char *p = NULL
            uint32_t map_size
        self.ensure_allocated(HEADER_CONST_LEN)

        map_size = 2 + <uint32_t>(schema_id > 0)

        p = begin = &self._buf[self._length]
        p = mp_encode_map(&p[5], map_size)
        p = mp_encode_uint(p, tarantool.IPROTO_REQUEST_TYPE)
        self.__op_offset = (p - begin)  # save op position
        p = mp_encode_uint(p, <uint32_t>op)
        p = mp_encode_uint(p, tarantool.IPROTO_SYNC)
        self.__sync_offset = (p - begin)  # save sync position
        p = mp_encode_uint(p, sync)

        if schema_id > 0:
            p = mp_encode_uint(p, tarantool.IPROTO_SCHEMA_VERSION)
            p = mp_store_u8(p, 0xce)
            self.__schema_id_offset = (p - begin)  # save schema_id position
            p = mp_store_u32(p, schema_id)

        self._length += (p - begin)

    cdef void change_schema_id(self, int64_t new_schema_id):
        cdef char *p

        if self.__schema_id_offset > 0 and new_schema_id > 0:
            p = &self._buf[self.__schema_id_offset]
            p = mp_store_u32(p, new_schema_id)

    cdef void write_length(self):
        cdef:
            char *p
        p = self._buf
        p = mp_store_u8(p, 0xce)
        p = mp_store_u32(p, self._length - 5)

    cdef char *_encode_nil(self, char *p) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, 1)
        p = mp_encode_nil(p)
        self._length += (p - begin)
        return p

    cdef char *_encode_bool(self, char *p, bint value) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, 1)
        p = mp_encode_bool(p, value)
        self._length += (p - begin)
        return p

    cdef char *_encode_double(self, char *p, double value) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, mp_sizeof_double(value))
        p = mp_encode_double(p, value)
        self._length += (p - begin)
        return p

    cdef char *_encode_uint(self, char *p, uint64_t value) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, mp_sizeof_uint(value))
        p = mp_encode_uint(p, value)
        self._length += (p - begin)
        return p

    cdef char *_encode_int(self, char *p, int64_t value) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, mp_sizeof_int(value))
        p = mp_encode_int(p, value)
        self._length += (p - begin)
        return p

    cdef char *_encode_str(self, char *p,
                           const char *str, uint32_t len) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, mp_sizeof_str(len))
        p = mp_encode_str(p, str, len)
        self._length += (p - begin)
        return p

    cdef char *_encode_bin(self, char *p,
                           const char *data, uint32_t len) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, mp_sizeof_bin(len))
        p = mp_encode_bin(p, data, len)
        self._length += (p - begin)
        return p

    cdef char *_encode_array(self, char *p, uint32_t len) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, mp_sizeof_array(len))
        p = mp_encode_array(p, len)
        self._length += (p - begin)
        return p

    cdef char *_encode_map(self, char *p, uint32_t len) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, mp_sizeof_map(len))
        p = mp_encode_map(p, len)
        self._length += (p - begin)
        return p

    cdef char *_encode_list(self, char *p, list arr) except NULL:
        cdef:
            uint32_t arr_len
            PyObject *item_ptr
            object item

        if arr is not None:
            arr_len = <uint32_t>cpython.list.PyList_GET_SIZE(arr)
        else:
            arr_len = 0
        p = self._encode_array(p, arr_len)

        if arr_len > 0:
            for i in range(arr_len):
                item_ptr = cpython.list.PyList_GET_ITEM(arr, i)
                item = <object>item_ptr
                p = self._encode_obj(p, item)
        return p

    cdef char *_encode_tuple(self, char *p, tuple t) except NULL:
        cdef:
            uint32_t t_len
            PyObject *item_ptr
            object item

        if t is not None:
            t_len = <uint32_t>cpython.tuple.PyTuple_GET_SIZE(t)
        else:
            t_len = 0
        p = self._encode_array(p, t_len)

        if t_len > 0:
            for i in range(t_len):
                item_ptr = cpython.tuple.PyTuple_GET_ITEM(t, i)
                item = <object>item_ptr
                p = self._encode_obj(p, item)
        return p

    cdef char *_encode_dict(self, char *p, dict d) except NULL:
        cdef:
            uint32_t d_len
            PyObject *pkey
            PyObject *pvalue
            object key, value
            Py_ssize_t pos

        if d is not None:
            d_len = <uint32_t>cpython.dict.PyDict_Size(d)
        else:
            d_len = 0
        p = self._encode_map(p, d_len)

        pos = 0
        while cpython.dict.PyDict_Next(d, &pos, &pkey, &pvalue):
            key = <object>pkey
            value = <object>pvalue
            p = self._encode_obj(p, key)
            p = self._encode_obj(p, value)

        return p

    cdef char *_encode_key_sequence(self, char *p, t,
                                    TntFields fields=None,
                                    bint default_none=False) except NULL:
        if isinstance(t, list) or t is None:
            return self._encode_list(p, <list>t)
        elif isinstance(t, tuple):
            return self._encode_tuple(p, <tuple>t)
        elif isinstance(t, dict) and fields is not None:
            return self._encode_list(
                p, dict_to_list_fields(<dict>t, fields, default_none)
            )
        else:
            if fields is not None:
                msg = 'sequence must be either list, tuple or dict'
            else:
                msg = 'sequence must be either list or tuple'
            raise TypeError(
                '{}, got: {}'.format(msg, type(t))
            )

    cdef char *_encode_obj(self, char *p, object o) except NULL:
        cdef:
            bytes o_string_temp
            char *o_string_str
            ssize_t o_string_len

        if o is None:
            return self._encode_nil(p)

        elif isinstance(o, bool):
            return self._encode_bool(p, <bint>o)

        elif isinstance(o, float):
            return self._encode_double(p, <double>o)

        elif isinstance(o, int):
            if o >= 0:
                return self._encode_uint(p, <uint64_t>o)
            else:
                return self._encode_int(p, <int64_t>o)

        elif isinstance(o, bytes):
            o_string_str = NULL
            o_string_len = 0
            cpython.bytes.PyBytes_AsStringAndSize(o,
                                                  &o_string_str,
                                                  &o_string_len)
            return self._encode_bin(p, o_string_str, <uint32_t>o_string_len)

        elif isinstance(o, str):
            o_string_temp = encode_unicode_string(o, self._encoding)
            o_string_str = NULL
            o_string_len = 0
            cpython.bytes.PyBytes_AsStringAndSize(o_string_temp,
                                                  &o_string_str,
                                                  &o_string_len)

            p = self._encode_str(p, o_string_str, <uint32_t>o_string_len)
            return p

        elif isinstance(o, list):
            return self._encode_list(p, <list>o)

        elif isinstance(o, tuple):
            return self._encode_tuple(p, <tuple>o)

        elif isinstance(o, dict):
            return self._encode_dict(p, <dict>o)

        else:
            raise TypeError(
                'Type `{}` is not supported for encoding'.format(type(o)))

    cdef char *_encode_update_ops(self, char *p, list operations,
                                  SchemaSpace space) except NULL:
        cdef:
            char *begin
            uint32_t ops_len, op_len
            bytes str_temp
            char *str_c
            ssize_t str_len

            char *op_str_c
            ssize_t op_str_len
            char op

            uint32_t extra_length

            uint64_t field_no
            object field_no_obj

            uint32_t splice_position, splice_offset

        begin = NULL

        if operations is not None:
            ops_len = <uint32_t>cpython.list.PyList_GET_SIZE(operations)
        else:
            ops_len = 0

        p = self._encode_array(p, ops_len)
        if ops_len == 0:
            return p

        for operation in operations:
            if not isinstance(operation, (list, tuple)):
                raise TypeError(
                    'Single operation must be a tuple or list')

            op_len = <uint32_t>cpython.list.PyList_GET_SIZE(operation)
            if op_len < 3:
                raise IndexError(
                    'Operation length must be at least 3')

            op_type_str = operation[0]
            if isinstance(op_type_str, str):
                str_temp = encode_unicode_string(op_type_str, self._encoding)
            elif isinstance(op_type_str, bytes):
                str_temp = <bytes>op_type_str
            else:
                raise TypeError(
                    'Operation type must of a str or bytes type')

            field_no_obj = operation[1]
            if isinstance(field_no_obj, int):
                field_no = <uint64_t>field_no_obj
            elif isinstance(field_no_obj, str):
                if space.fields is not None:
                    field_no = <uint64_t>space.fields.id_by_name(field_no_obj)
                else:
                    raise TypeError(
                        'Operation field_no must be int as there is '
                        'no format declaration in space {}'.format(space.sid))
            else:
                raise TypeError(
                    'Operation field_no must be of either int or str type')

            cpython.bytes.PyBytes_AsStringAndSize(str_temp, &op_str_c,
                                                  &op_str_len)
            op = <char>0
            if op_str_len == 1:
                op = op_str_c[0]

            if op == tarantool.IPROTO_OP_ADD \
                    or op == tarantool.IPROTO_OP_SUB \
                    or op == tarantool.IPROTO_OP_AND \
                    or op == tarantool.IPROTO_OP_XOR \
                    or op == tarantool.IPROTO_OP_OR \
                    or op == tarantool.IPROTO_OP_DELETE:
                op_argument = operation[2]
                if not isinstance(op_argument, int):
                    raise TypeError(
                        'int argument required for '
                        'Arithmetic and Delete operations'
                    )
                # mp_sizeof_array(3)
                # + mp_sizeof_str(1)
                # + mp_sizeof_uint(field_no)
                extra_length = 1 + 2 + mp_sizeof_uint(field_no)
                p = begin = self._ensure_allocated(p, extra_length)

                p = mp_encode_array(p, 3)
                p = mp_encode_str(p, op_str_c, 1)
                p = mp_encode_uint(p, field_no)
                self._length += (p - begin)
                p = self._encode_obj(p, op_argument)
            elif op == tarantool.IPROTO_OP_INSERT \
                    or op == tarantool.IPROTO_OP_ASSIGN:
                op_argument = operation[2]

                # mp_sizeof_array(3)
                # + mp_sizeof_str(1)
                # + mp_sizeof_uint(field_no)
                extra_length = 1 + 2 + mp_sizeof_uint(field_no)
                p = begin = self._ensure_allocated(p, extra_length)

                p = mp_encode_array(p, 3)
                p = mp_encode_str(p, op_str_c, 1)
                p = mp_encode_uint(p, field_no)
                self._length += (p - begin)
                p = self._encode_obj(p, op_argument)

            elif op == tarantool.IPROTO_OP_SPLICE:
                if op_len < 5:
                    raise IndexError(
                        'Splice operation must have length of 5, '
                        'but got: {}'.format(op_len)
                    )

                splice_position_obj = operation[2]
                splice_offset_obj = operation[3]
                op_argument = operation[4]
                if not isinstance(splice_position_obj, int):
                    raise TypeError('Splice position must be int')
                if not isinstance(splice_offset_obj, int):
                    raise TypeError('Splice offset must be int')

                splice_position = <uint32_t>splice_position_obj
                splice_offset = <uint32_t>splice_offset_obj

                # mp_sizeof_array(5) + mp_sizeof_str(1) + ...
                extra_length = 1 + 2 \
                                + mp_sizeof_uint(field_no) \
                                + mp_sizeof_uint(splice_position) \
                                + mp_sizeof_uint(splice_offset)
                p = begin = self._ensure_allocated(p, extra_length)

                p = mp_encode_array(p, 5)
                p = mp_encode_str(p, op_str_c, 1)
                p = mp_encode_uint(p, field_no)
                p = mp_encode_uint(p, splice_position)
                p = mp_encode_uint(p, splice_offset)
                self._length += (p - begin)
                p = self._encode_obj(p, op_argument)
            else:
                raise TypeError(
                    'Unknown update operation type `{}`'.format(op_type_str))
        return p

    cdef void encode_request_call(self, str func_name, args) except *:
        cdef:
            char *begin
            char *p
            uint32_t body_map_sz
            uint32_t max_body_len

            bytes func_name_temp
            char *func_name_str
            ssize_t func_name_len

        func_name_str = NULL
        func_name_len = 0

        func_name_temp = encode_unicode_string(func_name, self._encoding)
        cpython.bytes.PyBytes_AsStringAndSize(func_name_temp,
                                              &func_name_str,
                                              &func_name_len)
        body_map_sz = 2
        # Size description:
        # mp_sizeof_map()
        # + mp_sizeof_uint(TP_FUNCTION)
        # + mp_sizeof_str(func_name)
        # + mp_sizeof_uint(TP_TUPLE)
        max_body_len = 1 \
                       + 1 \
                       + mp_sizeof_str(<uint32_t>func_name_len) \
                       + 1

        self.ensure_allocated(max_body_len)

        p = begin = &self._buf[self._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_FUNCTION_NAME)
        p = mp_encode_str(p, func_name_str, <uint32_t>func_name_len)

        p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
        self._length += (p - begin)
        p = self._encode_key_sequence(p, args)

    cdef void encode_request_eval(self, str expression, args) except *:
        cdef:
            char *begin
            char *p
            uint32_t body_map_sz
            uint32_t max_body_len

            bytes expression_temp
            char *expression_str
            ssize_t expression_len

        expression_str = NULL
        expression_len = 0

        expression_temp = encode_unicode_string(expression, self._encoding)
        cpython.bytes.PyBytes_AsStringAndSize(expression_temp,
                                              &expression_str,
                                              &expression_len)
        body_map_sz = 2
        # Size description:
        # mp_sizeof_map()
        # + mp_sizeof_uint(TP_EXPRESSION)
        # + mp_sizeof_str(expression)
        # + mp_sizeof_uint(TP_TUPLE)
        max_body_len = 1 \
                       + 1 \
                       + mp_sizeof_str(<uint32_t>expression_len) \
                       + 1

        self.ensure_allocated(max_body_len)

        p = begin = &self._buf[self._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_EXPR)
        p = mp_encode_str(p, expression_str, <uint32_t>expression_len)

        p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
        self._length += (p - begin)
        p = self._encode_key_sequence(p, args)

    cdef void encode_request_select(self, SchemaSpace space, SchemaIndex index,
                                    key, uint64_t offset, uint64_t limit,
                                    uint32_t iterator) except *:
        cdef:
            char *begin
            char *p
            uint32_t body_map_sz
            uint32_t max_body_len
            uint32_t space_id, index_id

        space_id = space.sid
        index_id = index.iid

        body_map_sz = 3 \
                      + <uint32_t>(index_id > 0) \
                      + <uint32_t>(offset > 0) \
                      + <uint32_t>(iterator > 0)
        # Size description:
        # mp_sizeof_map(body_map_sz)
        # + mp_sizeof_uint(TP_SPACE)
        # + mp_sizeof_uint(space)
        # + mp_sizeof_uint(TP_LIMIT)
        # + mp_sizeof_uint(limit)
        max_body_len = 1 \
                       + 1 \
                       + 9 \
                       + 1 \
                       + 9

        if index_id > 0:
            # mp_sizeof_uint(TP_INDEX) + mp_sizeof_uint(index_id)
            max_body_len += 1 + 9
        if offset > 0:
            # mp_sizeof_uint(TP_OFFSET) + mp_sizeof_uint(offset)
            max_body_len += 1 + 9
        if iterator > 0:
            # mp_sizeof_uint(TP_ITERATOR) + mp_sizeof_uint(iterator)
            max_body_len += 1 + 1

        max_body_len += 1  # mp_sizeof_uint(TP_KEY);

        self.ensure_allocated(max_body_len)

        p = begin = &self._buf[self._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
        p = mp_encode_uint(p, space_id)
        p = mp_encode_uint(p, tarantool.IPROTO_LIMIT)
        p = mp_encode_uint(p, limit)

        if index_id > 0:
            p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
            p = mp_encode_uint(p, index_id)
        if offset > 0:
            p = mp_encode_uint(p, tarantool.IPROTO_OFFSET)
            p = mp_encode_uint(p, offset)
        if iterator > 0:
            p = mp_encode_uint(p, tarantool.IPROTO_ITERATOR)
            p = mp_encode_uint(p, iterator)

        p = mp_encode_uint(p, tarantool.IPROTO_KEY)
        self._length += (p - begin)
        p = self._encode_key_sequence(p, key, index.fields, False)

    cdef void encode_request_insert(self, SchemaSpace space, t) except *:
        cdef:
            char *begin
            char *p
            uint32_t body_map_sz
            uint32_t max_body_len
            uint32_t space_id

        space_id = space.sid

        body_map_sz = 2
        # Size description:
        # mp_sizeof_map(body_map_sz)
        # + mp_sizeof_uint(TP_SPACE)
        # + mp_sizeof_uint(space)
        # + mp_sizeof_uint(TP_TUPLE)
        max_body_len = 1 \
                       + 1 \
                       + 9 \
                       + 1

        self.ensure_allocated(max_body_len)

        p = begin = &self._buf[self._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
        p = mp_encode_uint(p, space_id)

        p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
        self._length += (p - begin)
        p = self._encode_key_sequence(p, t, space.fields, True)

    cdef void encode_request_delete(self, SchemaSpace space, SchemaIndex index,
                                    key) except *:
        cdef:
            char *p
            uint32_t body_map_sz
            uint32_t max_body_len
            uint32_t space_id, index_id

        space_id = space.sid
        index_id = index.iid

        body_map_sz = 2 \
                      + <uint32_t>(index_id > 0)
        # Size description:
        # mp_sizeof_map(body_map_sz)
        # + mp_sizeof_uint(TP_SPACE)
        # + mp_sizeof_uint(space)
        max_body_len = 1 \
                       + 1 \
                       + 9

        if index_id > 0:
            # mp_sizeof_uint(TP_INDEX) + mp_sizeof_uint(index)
            max_body_len += 1 + 9

        max_body_len += 1  # mp_sizeof_uint(TP_KEY);

        self.ensure_allocated(max_body_len)

        p = begin = &self._buf[self._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
        p = mp_encode_uint(p, space_id)

        if index_id > 0:
            p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
            p = mp_encode_uint(p, index_id)

        p = mp_encode_uint(p, tarantool.IPROTO_KEY)
        self._length += (p - begin)
        p = self._encode_key_sequence(p, key, index.fields, False)

    cdef void encode_request_update(self, SchemaSpace space, SchemaIndex index,
                                    key_tuple, list operations,
                                    bint is_upsert=False) except *:
        cdef:
            char *begin
            char *p
            uint32_t body_map_sz
            uint32_t max_body_len
            uint32_t space_id, index_id
            uint32_t key_of_tuple, key_of_operations
            TntFields fields
            bint default_fields_none

        space_id = space.sid
        index_id = index.iid

        if not is_upsert:
            key_of_tuple = tarantool.IPROTO_KEY
            key_of_operations = tarantool.IPROTO_TUPLE
            fields = index.fields
            default_fields_none = False
        else:
            key_of_tuple = tarantool.IPROTO_TUPLE
            key_of_operations = tarantool.IPROTO_OPS
            fields = space.fields
            default_fields_none = True

        body_map_sz = 3 + <uint32_t>(index_id > 0)
        # Size description:
        # mp_sizeof_map(body_map_sz)
        # + mp_sizeof_uint(TP_SPACE)
        # + mp_sizeof_uint(space)
        max_body_len = 1 \
                       + 1 \
                       + 9

        if index_id > 0:
            # + mp_sizeof_uint(TP_INDEX)
            # + mp_sizeof_uint(index)
            max_body_len += 1 + 9

        max_body_len += 1  # + mp_sizeof_uint(TP_KEY)
        max_body_len += 1  # + mp_sizeof_uint(TP_TUPLE)

        self.ensure_allocated(max_body_len)

        p = begin = &self._buf[self._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_SPACE_ID)
        p = mp_encode_uint(p, space_id)

        if index_id > 0:
            p = mp_encode_uint(p, tarantool.IPROTO_INDEX_ID)
            p = mp_encode_uint(p, index_id)
        self._length += (p - begin)

        p = self._encode_uint(p, key_of_tuple)
        p = self._encode_key_sequence(p, key_tuple, fields,
                                      default_fields_none)

        p = self._encode_uint(p, key_of_operations)
        p = self._encode_update_ops(p, operations, space)

    cdef void encode_request_upsert(self, SchemaSpace space, t,
                                    list operations) except *:
        self.encode_request_update(space, space.get_index(0),
                                   t, operations, True)

    cdef void encode_request_sql(self, str query, args) except *:
        cdef:
            char *begin
            char *p
            uint32_t body_map_sz
            uint32_t max_body_len

            bytes query_temp
            char *query_str
            ssize_t query_len

        query_str = NULL
        query_len = 0

        query_temp = encode_unicode_string(query, self._encoding)
        cpython.bytes.PyBytes_AsStringAndSize(query_temp,
                                              &query_str,
                                              &query_len)
        body_map_sz = 2
        # Size description:
        # mp_sizeof_map()
        # + mp_sizeof_uint(TP_SQL_TEXT)
        # + mp_sizeof_str(query)
        # + mp_sizeof_uint(TP_SQL_BIND)
        max_body_len = 1 \
                       + 1 \
                       + mp_sizeof_str(<uint32_t>query_len) \
                       + 1

        self.ensure_allocated(max_body_len)

        p = begin = &self._buf[self._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_SQL_TEXT)
        p = mp_encode_str(p, query_str, <uint32_t>query_len)

        p = mp_encode_uint(p, tarantool.IPROTO_SQL_BIND)
        self._length += (p - begin)
        # TODO: replace with custom encoder
        # TODO: need to simultaneously encode ordinal and named params
        p = self._encode_key_sequence(p, args)

    cdef void encode_request_auth(self,
                                  bytes username,
                                  bytes scramble) except *:
        cdef:
            char *begin
            char *p
            uint32_t body_map_sz
            uint32_t max_body_len

            char *username_str
            ssize_t username_len

            char *scramble_str
            ssize_t scramble_len

        cpython.bytes.PyBytes_AsStringAndSize(username,
                                              &username_str, &username_len)
        cpython.bytes.PyBytes_AsStringAndSize(scramble,
                                              &scramble_str, &scramble_len)
        body_map_sz = 2
        # Size description:
        # mp_sizeof_map()
        # + mp_sizeof_uint(TP_USERNAME)
        # + mp_sizeof_str(username_len)
        # + mp_sizeof_uint(TP_TUPLE)
        # + mp_sizeof_array(2)
        # + mp_sizeof_str(9) (chap-sha1)
        # + mp_sizeof_str(SCRAMBLE_SIZE)
        max_body_len = 1 \
                       + 1 \
                       + mp_sizeof_str(<uint32_t>username_len) \
                       + 1 \
                       + 1 \
                       + 1 + 9 \
                       + mp_sizeof_str(<uint32_t>scramble_len)

        self.ensure_allocated(max_body_len)

        p = begin = &self._buf[self._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_USER_NAME)
        p = mp_encode_str(p, username_str, <uint32_t>username_len)

        p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
        p = mp_encode_array(p, 2)
        p = mp_encode_str(p, "chap-sha1", 9)
        p = mp_encode_str(p, scramble_str, <uint32_t>scramble_len)
        self._length += (p - begin)



from libc.stdint cimport uint64_t, uint32_t, int64_t, int32_t

cdef struct Header:
    int32_t code
    int32_t return_code
    uint64_t sync
    int64_t schema_id


cdef class Response:
    cdef:
        int32_t _code
        int32_t _return_code
        uint64_t _sync
        int64_t _schema_id
        str _errmsg
        int _rowcount
        list _body
        bytes _encoding
        TntFields _fields
        list _autoincrement_ids
        bint _push_subscribe
        object _exception

        readonly object _q
        readonly object _push_event
        object _q_append
        object _q_popleft
        object _push_event_set
        object _push_event_clear

    cdef inline bint is_error(self)
    cdef inline uint32_t _len(self)
    cdef inline void add_push(self, push)
    cdef inline object pop_push(self)
    cdef inline object push_len(self)
    cdef inline void set_data(self, list data)
    cdef inline void set_exception(self, exc)
    cdef inline object get_exception(self)
    cdef inline void notify(self)


cdef ssize_t response_parse_header(const char *buf, uint32_t buf_len,
                                   Header *hdr) except -1
cdef ssize_t response_parse_body(const char *buf, uint32_t buf_len,
                                 Response resp, Request req,
                                 bint is_chunk) except -1



from asynctnt import Response
from asynctnt.exceptions import TarantoolSchemaError
from tests import BaseTarantoolTestCase


class UpsertTestCase(BaseTarantoolTestCase):
    async def _fill_data(self):
        data = [
            [0, 'a', 1],
            [1, 'b', 0],
        ]
        for t in data:
            await self.conn.insert(self.TESTER_SPACE_ID, t)

        return data

    async def test__upsert_empty_one_assign(self):
        data = [0, 'hello2', 1, 4, 'what is up']

        res = await self.conn.upsert(self.TESTER_SPACE_ID,
                                     data, [['=', 2, 2]])
        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, [], 'Body ok')

        res = await self.conn.select(self.TESTER_SPACE_ID, [0])
        self.assertResponseEqual(res, [data], 'Body ok')

    async def test__upsert_update_one_assign(self):
        data = [0, 'hello2', 1, 4, 'what is up']

        await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = await self.conn.upsert(self.TESTER_SPACE_ID,
                                     data, [['=', 2, 2]])
        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, [], 'Body ok')

        res = await self.conn.select(self.TESTER_SPACE_ID, [0])
        data[2] = 2
        self.assertResponseEqual(res, [data], 'Body ok')

    async def test__upsert_by_name(self):
        data = [0, 'hello2', 1, 4, 'what is up']

        await self.conn.upsert(self.TESTER_SPACE_NAME,
                               data, [['=', 2, 2]])

        res = await self.conn.select(self.TESTER_SPACE_ID, [0])
        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, [data], 'Body ok')

    async def test__upsert_by_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.upsert(self.TESTER_SPACE_NAME,
                                   [0, 'hello', 1], [['=', 2, 2]])

    async def test__upsert_dict_key(self):
        data = {
            'f1': 0,
            'f2': 'hello',
            'f3': 1,
            'f4': 2,
            'f5': 100,
        }

        res = await self.conn.upsert(self.TESTER_SPACE_ID,
                                     data, [['=', 2, 2]])
        self.assertResponseEqual(res, [], 'Body ok')

        res = await self.conn.select(self.TESTER_SPACE_ID, [0])
        self.assertResponseEqual(res,
                                 [[0, 'hello', 1, 2, 100]],
                                 'Body ok')

    async def test__usert_dict_resp_no_effect(self):
        data = {
            'f1': 0,
            'f2': 'hello',
            'f3': 1,
            'f4': 10,
            'f5': 1000,
        }

        res = await self.conn.upsert(self.TESTER_SPACE_ID, data, [['=', 2, 2]])
        self.assertResponseEqual(res, [], 'Body ok')



# cython: language_level=3

cimport cpython.dict

import asyncio
import enum

from asynctnt.exceptions import TarantoolNotConnectedError

include "const.pxi"

include "unicodeutil.pyx"
include "schema.pyx"
include "buffer.pyx"
include "rbuffer.pyx"
include "request.pyx"
include "response.pyx"
include "db.pyx"
include "push.pyx"

include "coreproto.pyx"


class Iterator(enum.IntEnum):
    """
        Available Iterator types
    """
    EQ = 0
    REQ = 1
    ALL = 2
    LT = 3
    LE = 4
    GE = 5
    GT = 6
    BITS_ALL_SET = 7
    BITS_ANY_SET = 8
    BITS_ALL_NOT_SET = 9
    OVERLAPS = 10
    NEIGHBOR = 11


cdef class BaseProtocol(CoreProtocol):
    def __init__(self, host, port,
                 username, password,
                 fetch_schema,
                 auto_refetch_schema,
                 connected_fut,
                 on_connection_made, on_connection_lost,
                 loop,
                 request_timeout=None,
                 encoding=None,
                 initial_read_buffer_size=None):
        CoreProtocol.__init__(self, host, port, encoding,
                              initial_read_buffer_size)

        self.loop = loop

        self.username = username
        self.password = password
        self.fetch_schema = fetch_schema
        self.auto_refetch_schema = auto_refetch_schema
        self.request_timeout = request_timeout or 0

        self.connected_fut = connected_fut
        self.on_connection_made_cb = on_connection_made
        self.on_connection_lost_cb = on_connection_lost
        self._closing = False

        self._on_request_completed_cb = self._on_request_completed
        self._on_request_timeout_cb = self._on_request_timeout

        self._reqs = {}
        self._sync = 0
        self._schema_id = -1
        self._schema = Schema.__new__(Schema, self._schema_id)
        self._schema_fetch_in_progress = False
        self._refetch_schema_future = None
        self._db = self._create_db()

        try:
            self.create_future = self.loop.create_future
        except AttributeError:
            self.create_future = self._create_future_fallback

    def _create_future_fallback(self):  # pragma: nocover
        return asyncio.Future(loop=self.loop)

    @property
    def schema(self):
        return self._schema

    @property
    def schema_id(self):
        return self._schema_id

    cdef void _set_connection_ready(self):
        if not self.connected_fut.done():
            self.connected_fut.set_result(True)
            self.con_state = CONNECTION_FULL

    cdef void _set_connection_error(self, e):
        if not self.connected_fut.done():
            self.connected_fut.set_exception(e)
            self.con_state = CONNECTION_BAD

    cdef void _on_greeting_received(self):
        if self.username and self.password:
            self._do_auth(self.username, self.password)
        elif self.fetch_schema:
            self._refetch_schema()
        else:
            self._set_connection_ready()

    cdef void _on_response_received(self, const char *buf, uint32_t buf_len):
        cdef:
            PyObject *req_p
            Request req
            Header hdr
            bint is_chunk
            object waiter
            object sync_obj
            object err

            ssize_t length

        length = response_parse_header(buf, buf_len, &hdr)
        buf_len -= length
        buf = &buf[length]  # skip header

        sync_obj = <object> hdr.sync

        req_p = cpython.dict.PyDict_GetItem(self._reqs, sync_obj)
        if req_p is NULL:
            logger.warning('sync %d not found', hdr.sync)
            return

        is_chunk = (hdr.code == tarantool.IPROTO_CHUNK)

        req = <Request> req_p
        req.response._sync = hdr.sync
        req.response._schema_id = hdr.schema_id
        if not is_chunk:
            req.response._code = hdr.code
            req.response._return_code = hdr.return_code
            cpython.dict.PyDict_DelItem(self._reqs, sync_obj)
        else:
            if not req.push_subscribe:
                # skip request data as no one will be waiting for it
                mp_next(&buf)
                return

        err = None
        if buf != &buf[buf_len]:
            # has body
            try:
                response_parse_body(buf, buf_len, req.response, req, is_chunk)
            except Exception as e:
                err = e

        # refetch schema if it is changed
        if self.auto_refetch_schema \
                and req.response._schema_id > 0 \
                and req.response._schema_id != self._schema_id:
            self._refetch_schema()

        # returning result
        if is_chunk:
            return

        waiter = req.waiter
        if waiter is None or waiter.done():
            return

        if err is not None:
            req.response.set_exception(err)
            waiter.set_exception(err)
            return

        if req.response.is_error():
            err = TarantoolDatabaseError(req.response._return_code,
                                         req.response._errmsg)
            req.response.set_exception(err)
            waiter.set_exception(err)
            return

        waiter.set_result(req.response)

    cdef void _do_auth(self, str username, str password):
        # No extra error handling from Db.execute
        fut = self._db._auth(self.salt, username, password, 0, False)

        def on_authorized(f):
            if f.cancelled():
                self._set_connection_error(asyncio.futures.CancelledError())
                return
            e = f.exception()
            if not e:
                logger.debug('Tarantool[%s:%s] Authorized successfully',
                             self.host, self.port)

                if self.fetch_schema:
                    self._do_fetch_schema(None)
                else:
                    self._set_connection_ready()
            else:
                logger.error('Tarantool[%s:%s] Authorization failed: %s',
                             self.host, self.port, str(e))
                self._set_connection_error(e)

        fut.add_done_callback(on_authorized)

    cdef void _do_fetch_schema(self, object fut):
        if self._schema_fetch_in_progress:
            return

        self._schema_fetch_in_progress = True

        def on_fetch(f):
            self._schema_fetch_in_progress = False

            if f.cancelled():
                self._set_connection_error(asyncio.futures.CancelledError())
                return
            e = f.exception()
            if not e:
                spaces, indexes = f.result()
                logger.debug('Tarantool[%s:%s] Schema fetch succeeded. '
                             'Version: %d, Spaces: %d, Indexes: %d.',
                             self.host, self.port,
                             spaces.schema_id, len(spaces), len(indexes))
                try:
                    self._schema = Schema.parse(spaces.schema_id,
                                                spaces, indexes)
                except Exception as e:
                    logger.exception(e)
                    logger.error('Error happened while parsing schema. '
                                 'Space, fields and index names currently '
                                 'not working. Please file an issue at '
                                 'https://github.com/igorcoding/asynctnt')
                    self.auto_refetch_schema = False
                    self.fetch_schema = False
                    self._schema_id = -1
                    self._set_connection_ready()
                    if fut is not None:
                        fut.set_result(None)
                    return

                self._schema_id = self._schema.id
                self._set_connection_ready()
                if fut is not None:
                    fut.set_result(self._schema)
            else:
                logger.error('Tarantool[%s:%s] Schema fetch failed: %s',
                             self.host, self.port, str(e))
                if isinstance(e, asyncio.TimeoutError):
                    e = asyncio.TimeoutError('Schema fetch timeout')
                self._set_connection_error(e)
                if fut is not None:
                    fut.set_exception(e)

        self._schema_id = -1
        fut_vspace = self._db.select(SPACE_VSPACE, timeout=0)
        fut_vindex = self._db.select(SPACE_VINDEX, timeout=0)
        gather_fut = asyncio.gather(fut_vspace, fut_vindex,
                                    return_exceptions=False)
        gather_fut.add_done_callback(on_fetch)

    cdef void _on_connection_made(self):
        CoreProtocol._on_connection_made(self)

        if self.on_connection_made_cb:
            self.on_connection_made_cb()

    cdef void _on_connection_lost(self, exc):
        cdef:
            Request req
            PyObject *pkey
            PyObject *pvalue
            object key, value
            Py_ssize_t pos

        if self._closing:
            return

        self._closing = True

        pos = 0
        while cpython.dict.PyDict_Next(self._reqs, &pos, &pkey, &pvalue):
            sync = <uint64_t> <object> pkey
            req = <Request> pvalue

            waiter = req.waiter
            if waiter and not waiter.done():
                err = None
                if exc is None:
                    err = TarantoolNotConnectedError(
                        'Lost connection to Tarantool'
                    )
                elif isinstance(exc, (ConnectionRefusedError,
                                      ConnectionResetError)):
                    err = TarantoolNotConnectedError(
                        'Lost connection to Tarantool: {}: {}'.format(
                            exc.__class__.__name__, str(exc))
                    )
                elif exc is ConnectionRefusedError \
                        or exc is ConnectionResetError:
                    err = TarantoolNotConnectedError(
                        'Lost connection to Tarantool: {}'.format(
                            exc.__name__)
                    )
                else:
                    err = exc

                if err is not None:
                    waiter.set_exception(err)
                    if req.response is not None:
                        req.response.set_exception(err)

        if self.on_connection_lost_cb:
            self.on_connection_lost_cb(exc)

        self._reqs = {}  # reset requests map

    cdef inline uint64_t next_sync(self):
        self._sync += 1
        return self._sync

    def _on_request_timeout(self, waiter):
        cdef Request req

        if waiter.done():
            return

        req = waiter._req
        req.timeout_handle.cancel()
        req.timeout_handle = None
        waiter.set_exception(
            asyncio.TimeoutError(
                '{} exceeded timeout'.format(req.__class__.__name__))
        )

    def _on_request_completed(self, fut):
        cdef Request req = fut._req
        fut._req = None

        if req.timeout_handle is not None:
            req.timeout_handle.cancel()
            req.timeout_handle = None

    cdef object _new_waiter_for_request(self, Request req, float timeout):
        fut = self.create_future()
        fut._req = req  # to be able to retrieve request after done()
        req.waiter = fut
        req.response = Response.__new__(Response, self.encoding,
                                        req.push_subscribe)

        if timeout < 0:
            timeout = self.request_timeout
        if timeout is not None and timeout > 0:
            req.timeout_handle = \
                self.loop.call_later(timeout,
                                     self._on_request_timeout_cb, fut)
            fut.add_done_callback(self._on_request_completed_cb)
        return fut

    cdef Db _create_db(self):
        return Db.new(self)

    def create_db(self):
        return self._create_db()

    def get_common_db(self):
        return self._db

    cdef object execute(self, Request req, WriteBuffer buf, float timeout):
        if self.con_state == CONNECTION_BAD:
            raise TarantoolNotConnectedError('Tarantool is not connected')

        cpython.dict.PyDict_SetItem(self._reqs, req.sync, req)
        self._write(buf)

        return self._new_waiter_for_request(req, timeout)

    cdef uint32_t transform_iterator(self, iterator) except *:
        if isinstance(iterator, int):
            return iterator
        if isinstance(iterator, Iterator):
            return iterator.value
        if isinstance(iterator, str):
            return Iterator[iterator]
        else:
            raise TypeError('Iterator is of unsupported type '
                            '(asynctnt.Iterator, int, str)')

    cdef object _refetch_schema(self):
        if self._refetch_schema_future is None \
                or self._refetch_schema_future.done():
            self._refetch_schema_future = self.create_future()
            self._do_fetch_schema(self._refetch_schema_future)

        return self._refetch_schema_future

    def refetch_schema(self):
        return self._refetch_schema()


class Protocol(BaseProtocol, asyncio.Protocol):
    pass


TarantoolTuple = <object> tupleobj.AtntTuple_InitTypes()



import logging

from asynctnt import Iterator
from asynctnt import Response
from asynctnt.exceptions import TarantoolSchemaError
from tests import BaseTarantoolTestCase
from tests.util import get_complex_param


class SelectTestCase(BaseTarantoolTestCase):
    LOGGING_LEVEL = logging.INFO

    async def _fill_data(self, count=3, space=None):
        space = space or self.TESTER_SPACE_ID
        data = []
        for i in range(count):
            t = [i, str(i), 1, 2, 'something']
            data.append(t)
            await self.conn.insert(space, t)
        return data

    async def _fill_data_dict(self, count=3):
        data = []
        for i in range(count):
            t = {
                'f1': i,
                'f2': str(i),
                'f3': 1,
                'f4': 2,
                'f5': 'something',
            }
            t = await self.conn.insert(self.TESTER_SPACE_ID, t)
            data.append(dict(t[0]))
        return data

    async def test__select_by_id_empty_space(self):
        res = await self.conn.select(self.TESTER_SPACE_ID)

        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, [], 'Body ok')

    async def test__select_by_id_non_empty_space(self):
        data = await self._fill_data()

        res = await self.conn.select(self.TESTER_SPACE_ID)

        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, data, 'Body ok')

    async def test__select_by_name_space_empty(self):
        res = await self.conn.select(self.TESTER_SPACE_NAME)

        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, [], 'Body ok')

    async def test__select_by_name_non_empty_space(self):
        data = await self._fill_data()

        res = await self.conn.select(self.TESTER_SPACE_NAME)

        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, data, 'Body ok')

    async def test__select_by_index_id(self):
        data = await self._fill_data()

        res = await self.conn.select(self.TESTER_SPACE_ID, index=1)

        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, data, 'Body ok')

    async def test__select_by_index_name(self):
        data = await self._fill_data()

        res = await self.conn.select(self.TESTER_SPACE_ID, index='txt')

        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, data, 'Body ok')

    async def test__select_by_id_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        try:
            await self.conn.select(self.TESTER_SPACE_ID)
        except Exception as e:
            self.fail(e)

    async def test__select_by_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.select(self.TESTER_SPACE_NAME)

    async def test__select_by_index_id_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        try:
            await self.conn.select(self.TESTER_SPACE_ID, index=1)
        except Exception as e:
            self.fail(e)

    async def test__select_by_index_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.select(self.TESTER_SPACE_NAME, index='txt')

    async def test__select_by_key_one_item(self):
        data = await self._fill_data()

        res = await self.conn.select(self.TESTER_SPACE_NAME, [1])
        self.assertResponseEqual(res, [data[1]], 'Body ok')

    async def test__select_by_key_multiple_items_index(self):
        data = await self._fill_data()
        next_id = data[-1][0] + 1
        next_txt = data[-1][1]
        await self.conn.insert(self.TESTER_SPACE_ID,
                               [next_id, next_txt, 1, 2, 'text'])
        data.append([next_id, next_txt, 1, 2, 'text'])

        res = await self.conn.select(self.TESTER_SPACE_NAME, [next_txt],
                                     index='txt')
        self.assertResponseEqual(res, data[len(data)-2:], 'Body ok')

    async def test__select_limit(self):
        data = await self._fill_data()

        res = await self.conn.select(self.TESTER_SPACE_NAME, limit=1)
        self.assertResponseEqual(res, [data[0]], 'Body ok')

    async def test__select_limit_offset(self):
        data = await self._fill_data(4)

        res = await self.conn.select(self.TESTER_SPACE_NAME,
                                     limit=1, offset=2)
        self.assertResponseEqual(res, [data[2]], 'Body ok')

    async def test__select_iterator_class(self):
        data = await self._fill_data(4)

        res = await self.conn.select(self.TESTER_SPACE_NAME,
                                     iterator=Iterator.GE)
        self.assertResponseEqual(res, data, 'Body ok')

        res = await self.conn.select(self.TESTER_SPACE_NAME,
                                     iterator=Iterator.LE)
        self.assertResponseEqual(res, list(reversed(data)), 'Body ok')

    async def test__select_iterator_int(self):
        data = await self._fill_data(4)

        res = await self.conn.select(self.TESTER_SPACE_NAME,
                                     iterator=4)
        self.assertResponseEqual(res, list(reversed(data)), 'Body ok')

    async def test__select_iterator_str(self):
        data = await self._fill_data(4)

        res = await self.conn.select(self.TESTER_SPACE_NAME,
                                     iterator='LE')
        self.assertResponseEqual(res, list(reversed(data)), 'Body ok')

    async def test__select_complex(self):
        p, p_cmp = get_complex_param(replace_bin=False)
        data = [1, 'hello2', 1, 4, p_cmp]

        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.select(self.TESTER_SPACE_ID)
        self.assertResponseEqual(res, [data], 'Body ok')

    async def test__select_all_params(self):
        data = await self._fill_data(10)

        res = await self.conn.select(self.TESTER_SPACE_NAME,
                                     index='primary',
                                     limit=2, offset=1,
                                     iterator=Iterator.LE)
        self.assertResponseEqual(res, list(reversed(data))[1:3], 'Body ok')

    async def test__select_all_by_hash_index(self):
        data = await self._fill_data(4, space='no_schema_space')

        res = await self.conn.select('no_schema_space',
                                     index='primary_hash')
        self.assertResponseEqual(res, data, 'Body ok')

    async def test__select_key_tuple(self):
        try:
            await self.conn.select(self.TESTER_SPACE_ID, (1,))
        except Exception as e:
            self.fail(e)

    async def test__select_invalid_types(self):
        with self.assertRaisesRegex(
                TypeError,
                r'missing 1 required positional argument: \'space\''):
            await self.conn.select()

        with self.assertRaisesRegex(
                TypeError,
                r'sequence must be either list, tuple or dict'):
            await self.conn.select(self.TESTER_SPACE_ID, 1)

        with self.assertRaisesRegex(
                TypeError, r'Index must be either str or int, got'):
            await self.conn.select(self.TESTER_SPACE_ID, [1],
                                   index=[1, 2])

        with self.assertRaisesRegex(
                TypeError, r'an integer is required'):
            await self.conn.select(self.TESTER_SPACE_ID, [1],
                                   index=1, limit='hello')

        with self.assertRaisesRegex(
                TypeError, r'an integer is required'):
            await self.conn.select(self.TESTER_SPACE_ID, [1],
                                   index=1, limit=1, offset='hello')

        with self.assertRaisesRegex(
                TypeError, r'Iterator is of unsupported type'):
            await self.conn.select(self.TESTER_SPACE_ID, [1],
                                   index=1, limit=1, offset=1,
                                   iterator=[1, 2])

    async def test__select_dict_key(self):
        data = await self._fill_data()
        res = await self.conn.select(self.TESTER_SPACE_ID, {
            'f1': data[0][0]
        })
        self.assertResponseEqual(res, [data[0]], 'Body ok')

    async def test__select_dict_key_wrong_field(self):
        data = await self._fill_data()
        res = await self.conn.select(self.TESTER_SPACE_ID, {
            'f2': data[0][0]
        })
        self.assertResponseEqual(res, data, 'Body ok')

    async def test__select_dict_key_other_index(self):
        data = await self._fill_data()
        res = await self.conn.select(self.TESTER_SPACE_ID, {
            'f2': data[0][1]
        }, index='txt')
        self.assertResponseEqual(res, [data[0]], 'Body ok')

    async def test__select_dict_resp(self):
        data = await self._fill_data_dict()
        res = await self.conn.select(self.TESTER_SPACE_ID, [])
        self.assertResponseEqualKV(res, data)



import warnings

from asynctnt import TarantoolTuple
from tests import BaseTarantoolTestCase


class ResponseTestCase(BaseTarantoolTestCase):
    async def _fill_data(self, count=3):
        data = []
        for i in range(count):
            t = [i, str(i), 1, 2, 'something']
            data.append(t)
            await self.conn.insert(self.TESTER_SPACE_ID, t)
        return data

    async def _fill_data_dict(self, count=3):
        data = []
        for i in range(count):
            t = {
                'f1': i,
                'f2': str(i),
                'f3': 1,
                'f4': 2,
                'f5': 'something',
            }
            t = await self.conn.insert(self.TESTER_SPACE_ID, t)
            data.append(dict(t[0]))
        return data

    async def test__response_indexing(self):
        res = await self.conn.call('box.info')
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            self.assertEqual(len(res), len(res.body), 'len ok')
            self.assertEqual(res[0], res.body[0], 'value ok')

    async def test__response_iter(self):
        data = await self._fill_data(3)
        res = await self.conn.select(self.TESTER_SPACE_ID)

        self.assertEqual(len(res), len(data), 'len ok')

        res_arr = []
        for el in res:
            res_arr.append(list(el))
        self.assertListEqual(res_arr, data, 'list ok')

    async def test__response_tuple_iter(self):
        data = await self._fill_data(1)
        res = await self.conn.select(self.TESTER_SPACE_ID)
        t = res[0]

        t_list = [el for el in t]  # check iteration over tuple
        self.assertListEqual(t_list, data[0], 'tuple ok')

    async def test__response_tuple_keys(self):
        await self._fill_data(1)
        res = await self.conn.select(self.TESTER_SPACE_ID)
        t = res[0]

        correct_keys = ['f1', 'f2', 'f3', 'f4', 'f5']
        self.assertListEqual(list(t.keys()), correct_keys, 'keys ok')

    async def test__response_tuple_values(self):
        data = await self._fill_data(1)
        res = await self.conn.select(self.TESTER_SPACE_ID)
        t = res[0]

        self.assertListEqual(list(t.values()), data[0], 'values ok')

    async def test__response_tuple_items(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        t = res[0]
        d = {
            'f1': data[0],
            'f2': data[1],
            'f3': data[2],
            'f4': data[3],
            'f5': data[4],
        }

        t_dict = {k: v for k, v in t.items()}
        self.assertDictEqual(t_dict, d, 'items ok')

    async def test__response_tuple_dict_extra_index(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        self.assertEqual(res[0], data[0])
        self.assertEqual(res[1], data[1])
        self.assertEqual(res[2], data[2])
        self.assertEqual(res[3], data[3])
        self.assertEqual(res[4], data[4])
        self.assertEqual(res[5], data[5])
        self.assertEqual(res[6], data[6])
        self.assertEqual(res[-1], data[-1])
        self.assertEqual(res[-3], data[-3])

        self.assertEqual(res['f1'], data[0])
        self.assertEqual(res['f2'], data[1])
        self.assertEqual(res['f3'], data[2])
        self.assertEqual(res['f4'], data[3])
        self.assertEqual(res['f5'], data[4])

    async def test__response_tuple_slice(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        self.assertEqual(type(res[:3]), tuple)

        self.assertListEqual(list(res[:3]), data[:3])
        self.assertListEqual(list(res[1:5]), data[1:5])
        self.assertListEqual(list(res[5:20]), data[5:20])
        self.assertEqual(list(res[7:3:-1]), data[7:3:-1])
        self.assertEqual(list(res[7:3:-2]), data[7:3:-2])

    async def test__response_tuple_contains(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        self.assertTrue('f1' in res)
        self.assertTrue('f2' in res)
        self.assertTrue('f3' in res)
        self.assertTrue('f4' in res)
        self.assertTrue('f5' in res)
        self.assertFalse('f6' in res)

    async def test__response_tuple_key_error(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        with self.assertRaises(KeyError):
            # noinspection PyStatementEffect
            res['f100']

    async def test__response_tuple_get(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        self.assertEqual(res.get('f1'), 0)
        self.assertEqual(res.get('f2'), 'hello')
        self.assertEqual(res.get('f100'), None)
        self.assertEqual(res.get('f100', 'zz'), 'zz')

    async def test__response_with_no_space_format(self):
        res = await self.conn.insert('no_schema_space', [0, 'one'])

        with self.assertRaises(ValueError):
            res[0].keys()

        with self.assertRaises(ValueError):
            res[0].items()

    async def test__native_response_with_no_space_format(self):
        await self.conn.insert('no_schema_space', [0, 'one'])

        res = await self.conn.select('no_schema_space')
        self.assertEqual(1, res.rowcount, 'count correct')
        self.assertTrue(isinstance(res[0], TarantoolTuple),
                        'expecting a TarantoolTuple')
        self.assertResponseEqual(res, [[0, 'one']], 'resp ok')

        try:
            repr(res)
        except Exception as e:
            self.fail(e)

    async def test__response_repr(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.select('tester')
        self.assertEqual(1, res.rowcount, 'count correct')
        self.assertTrue(isinstance(res[0], TarantoolTuple),
                        'expecting a TarantoolTuple')

        self.assertEqual(
            "<TarantoolTuple f1=0 f2='hello' f3=5 f4=6 f5='help' "
            "5='common' 6='yo'>",
            repr(res[0]), 'repr ok')

    async def test__response_str(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.select('tester')
        self.assertEqual(1, res.rowcount, 'count correct')
        self.assertTrue(isinstance(res[0], TarantoolTuple),
                        'expecting a TarantoolTuple')

        try:
            str(res)
        except Exception as e:
            self.fail(e)



import asyncio

import asynctnt
from asynctnt._testbase import ensure_version
from asynctnt.exceptions import TarantoolNotConnectedError, \
    TarantoolDatabaseError

from tests import BaseTarantoolTestCase
from tests.util import get_complex_param, get_big_param


class CommonTestCase(BaseTarantoolTestCase):
    async def test__encoding_utf8(self):
        p, p_cmp = get_complex_param(replace_bin=False)

        data = [1, 'hello', 1, 0, p]
        data_cmp = [1, 'hello', 1, 0, p_cmp]

        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        self.assertResponseEqual(res, [data_cmp], 'Body ok')

        res = await self.conn.select(self.TESTER_SPACE_ID)
        self.assertResponseEqual(res, [data_cmp], 'Body ok')

    async def test__encoding_cp1251(self):
        await self.tnt_reconnect(encoding='cp1251')
        p, p_cmp = get_complex_param(replace_bin=False)

        data = [1, 'hello', 1, 0, p]
        data_cmp = [1, 'hello', 1, 0, p_cmp]

        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        self.assertResponseEqual(res, [data_cmp], 'Body ok')

        res = await self.conn.select(self.TESTER_SPACE_ID)
        self.assertResponseEqual(res, [data_cmp], 'Body ok')

    @ensure_version(max=(2, 0), max_included=True)
    async def test__schema_refetch_on_schema_change(self):
        await self.tnt_reconnect(auto_refetch_schema=True,
                                 username='t1', password='t1')
        self.assertTrue(self.conn.fetch_schema)
        self.assertTrue(self.conn.auto_refetch_schema)
        schema_before = self.conn.schema_id
        self.assertNotEqual(schema_before, -1)

        # Changing scheme
        await self.conn.eval(
            "local s = box.schema.create_space('new_space');"
            "s:drop();"
        )

        try:
            await self.conn.ping()
        except Exception as e:
            self.fail(e)

        # wait for schema to refetch
        await self.sleep(1)

        self.assertGreater(self.conn.schema_id, schema_before,
                           'Schema changed')

    async def test__schema_refetch_manual(self):
        await self.tnt_reconnect(fetch_schema=True,
                                 auto_refetch_schema=False,
                                 username='t1', password='t1')
        self.assertTrue(self.conn.fetch_schema)
        self.assertFalse(self.conn.auto_refetch_schema)
        schema_before = self.conn.schema_id
        self.assertNotEqual(schema_before, -1)

        await self.conn.call('change_format')

        try:
            await self.conn.ping()
        except Exception as e:
            self.fail(e)

        self.assertEqual(self.conn.schema_id, schema_before,
                         'schema not changed')

        await self.conn.refetch_schema()

        self.assertGreater(self.conn.schema_id, schema_before,
                           'Schema changed')

    @ensure_version(max=(2, 0), max_included=True)
    async def test__schema_no_fetch_and_refetch(self):
        await self.tnt_reconnect(auto_refetch_schema=False,
                                 username='t1', password='t1',
                                 fetch_schema=False)
        self.assertFalse(self.conn.fetch_schema)
        self.assertFalse(self.conn.auto_refetch_schema)
        self.assertEqual(self.conn.schema_id, -1)

        # Changing scheme
        await self.conn.eval(
            "s = box.schema.create_space('new_space');"
            "s:drop();"
        )

        try:
            await self.conn.ping()
        except Exception as e:
            self.fail(e)

    async def test__parse_numeric_map_keys(self):
        res = await self.conn.eval(
            """return {
                [1] = 1,
                [2] = 2,
                hello = 3,
                world = 4,
                [-3] = 5,
                [4.5] = 6
            }"""
        )

        d = {
            1: 1,
            2: 2,
            'hello': 3,
            'world': 4,
            -3: 5,
            4.5: 6
        }

        self.assertDictEqual(res[0], d, 'Numeric keys parsed ok')

    async def test__read_buffer_reallocate_ok(self):
        await self.tnt_reconnect(initial_read_buffer_size=1)

        p, cmp = get_complex_param(encoding=self.conn.encoding)
        try:
            res = await self.conn.call('func_param', [p])
        except Exception as e:
            self.fail(e)

        self.assertDictEqual(res[0][0], cmp, 'Body ok')

    async def test__read_buffer_deallocate_ok(self):
        size = 100 * 1000
        await self.tnt_reconnect(initial_read_buffer_size=size)

        # Waiting big response, so ReadBuffer grows to hold it
        p = get_big_param(size=size * 3)
        try:
            await self.conn.call('func_param', [p])
        except Exception as e:
            self.fail(e)

        # Waiting small response, so ReadBuffer deallocates memory
        p = get_big_param(size=10)
        try:
            await self.conn.call('func_param', [p])
        except Exception as e:
            self.fail(e)

    async def test__write_buffer_reallocate(self):
        p = get_big_param(size=100*1024)
        try:
            res = await self.conn.call('func_param', [p])
        except Exception as e:
            self.fail(e)

        self.assertDictEqual(res[0][0], p, 'Body ok')

    async def test__ensure_no_attribute_error_on_not_connected(self):
        await self.tnt_disconnect()

        self._conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port)

        with self.assertRaises(TarantoolNotConnectedError):
            await self.conn.ping()

    async def test__encode_unsupported_type(self):
        class A:
            pass
        with self.assertRaisesRegex(
                TypeError, 'Type `(.+)` is not supported for encoding'):
            await self.conn.call('func_param', [{'a': A()}])

    async def test__schema_refetch_next_byte(self):
        await self.tnt_reconnect(auto_refetch_schema=True,
                                 username='t1', password='t1')
        await self.conn.call('func_hello')

        # Changing scheme
        try:
            for _ in range(251):
                await self.conn.eval(
                    "s = box.schema.create_space('new_space');"
                    "s:drop();"
                )
        except TarantoolDatabaseError as e:
            self.fail(e)

        try:
            for i in range(1, 255):
                await self.conn.call('func_hello')
        except TarantoolDatabaseError as e:
            self.fail(e)

    async def test__schema_refetch_unknown_space(self):
        await self.tnt_reconnect(auto_refetch_schema=True,
                                 username='t1', password='t1',
                                 ping_timeout=0.1)

        async def func():
            # trying to select from an unknown space until it is created
            while True:
                try:
                    await self.conn.select('spacex')
                    return
                except Exception:
                    pass

                await asyncio.sleep(0.1)

        f = asyncio.ensure_future(asyncio.wait_for(func(), timeout=1))

        # Changing scheme
        try:
            conn = await asynctnt.connect(host=self.tnt.host,
                                          port=self.tnt.port,
                                          username='t1', password='t1')
            async with conn:
                await conn.eval(
                    "s = box.schema.create_space('spacex');"
                    "s:create_index('primary');"
                )
        except TarantoolDatabaseError as e:
            self.fail(e)

        try:
            await f
        except (asyncio.TimeoutError, asyncio.CancelledError) as e:
            self.fail('Schema is not updated: %s %s' % (type(e), e))



from libc.stdint cimport uint32_t, uint64_t, int64_t


cdef class Request:
    cdef:
        tarantool.iproto_type op
        uint64_t sync
        int64_t schema_id
        SchemaSpace space
        object waiter
        object timeout_handle
        bint parse_metadata
        bint parse_as_tuples
        bint push_subscribe
        Response response

    @staticmethod
    cdef inline Request new(tarantool.iproto_type op,
                            uint64_t sync, int64_t schema_id,
                            SchemaSpace space, bint push_subscribe)

    cdef inline TntFields fields(self):
        if self.space is None:
            return None
        return self.space.fields



cimport cpython

import hashlib

cdef class Db:
    def __cinit__(self):
        self._protocol = None
        self._encoding = None

    @staticmethod
    cdef inline Db new(BaseProtocol protocol):
        cdef Db db = Db.__new__(Db)
        db._protocol = protocol
        db._encoding = protocol.encoding
        return db

    cdef inline uint64_t next_sync(self):
        return self._protocol.next_sync()

    cdef object _ping(self, float timeout, bint push_subscribe):
        cdef:
            tarantool.iproto_type op
            uint64_t sync
            int64_t schema_id
            WriteBuffer buf

        op = tarantool.IPROTO_PING
        sync = self.next_sync()
        schema_id = -1  # not sending schema_id with the request
        buf = WriteBuffer.new(self._encoding)
        buf.write_header(sync, op, schema_id)
        buf.write_length()
        return self._protocol.execute(
            Request.new(op, sync, schema_id, None, push_subscribe),
            buf, timeout)

    cdef object _call16(self, str func_name, args,
                         float timeout, bint push_subscribe):
        cdef:
            tarantool.iproto_type op
            uint64_t sync
            int64_t schema_id
            WriteBuffer buf

        op = tarantool.IPROTO_CALL_16
        sync = self.next_sync()
        schema_id = -1  # not sending schema_id with the request
        buf = WriteBuffer.new(self._encoding)
        buf.write_header(sync, op, schema_id)
        buf.encode_request_call(func_name, args)
        buf.write_length()
        return self._protocol.execute(
            Request.new(op, sync, schema_id, None, push_subscribe),
            buf, timeout)

    cdef object _call(self, str func_name, args,
                         float timeout, bint push_subscribe):
        cdef:
            tarantool.iproto_type op
            uint64_t sync
            int64_t schema_id
            WriteBuffer buf

        op = tarantool.IPROTO_CALL
        sync = self.next_sync()
        schema_id = -1  # not sending schema_id with the request
        buf = WriteBuffer.new(self._encoding)
        buf.write_header(sync, op, schema_id)
        buf.encode_request_call(func_name, args)
        buf.write_length()
        return self._protocol.execute(
            Request.new(op, sync, schema_id, None, push_subscribe),
            buf, timeout)

    cdef object _eval(self, str expression, args,
                         float timeout, bint push_subscribe):
        cdef:
            tarantool.iproto_type op
            uint64_t sync
            int64_t schema_id
            WriteBuffer buf

        op = tarantool.IPROTO_EVAL
        sync = self.next_sync()
        schema_id = -1  # not sending schema_id with the request
        buf = WriteBuffer.new(self._encoding)
        buf.write_header(sync, op, schema_id)
        buf.encode_request_eval(expression, args)
        buf.write_length()
        return self._protocol.execute(
            Request.new(op, sync, schema_id, None, push_subscribe),
            buf, timeout)

    cdef object _select(self, SchemaSpace space, SchemaIndex index, key,
                         uint64_t offset, uint64_t limit, uint32_t iterator,
                         float timeout, bint push_subscribe):
        cdef:
            tarantool.iproto_type op
            uint64_t sync
            int64_t schema_id
            WriteBuffer buf

        op = tarantool.IPROTO_SELECT
        sync = self.next_sync()
        schema_id = -1  # not sending schema_id with the request
        buf = WriteBuffer.new(self._encoding)
        buf.write_header(sync, op, schema_id)
        buf.encode_request_select(space, index, key,
                                  offset, limit, iterator)
        buf.write_length()
        return self._protocol.execute(
            Request.new(op, sync, schema_id, space, push_subscribe),
            buf, timeout)

    cdef object _insert(self, SchemaSpace space, t, bint replace,
                         float timeout, bint push_subscribe):
        cdef:
            tarantool.iproto_type op
            uint64_t sync
            int64_t schema_id
            WriteBuffer buf

        op = tarantool.IPROTO_REPLACE if replace else tarantool.IPROTO_INSERT
        sync = self.next_sync()
        schema_id = -1  # not sending schema_id with the request
        buf = WriteBuffer.new(self._encoding)
        buf.write_header(sync, op, schema_id)
        buf.encode_request_insert(space, t)
        buf.write_length()
        return self._protocol.execute(
            Request.new(op, sync, schema_id, space, push_subscribe),
            buf, timeout)

    cdef object _delete(self, SchemaSpace space, SchemaIndex index, key,
                         float timeout, bint push_subscribe):
        cdef:
            tarantool.iproto_type op
            uint64_t sync
            int64_t schema_id
            WriteBuffer buf

        op = tarantool.IPROTO_DELETE
        sync = self.next_sync()
        schema_id = -1  # not sending schema_id with the request
        buf = WriteBuffer.new(self._encoding)
        buf.write_header(sync, op, schema_id)
        buf.encode_request_delete(space, index, key)
        buf.write_length()
        return self._protocol.execute(
            Request.new(op, sync, schema_id, space, push_subscribe),
            buf, timeout)

    cdef object _update(self, SchemaSpace space, SchemaIndex index,
                         key, list operations,
                         float timeout, bint push_subscribe):
        cdef:
            tarantool.iproto_type op
            uint64_t sync
            int64_t schema_id
            WriteBuffer buf

        op = tarantool.IPROTO_UPDATE
        sync = self.next_sync()
        schema_id = -1  # not sending schema_id with the request
        buf = WriteBuffer.new(self._encoding)
        buf.write_header(sync, op, schema_id)
        buf.encode_request_update(space, index, key, operations)
        buf.write_length()
        return self._protocol.execute(
            Request.new(op, sync, schema_id, space, push_subscribe),
            buf, timeout)

    cdef object _upsert(self, SchemaSpace space, t, list operations,
                         float timeout, bint push_subscribe):
        cdef:
            tarantool.iproto_type op
            uint64_t sync
            int64_t schema_id
            WriteBuffer buf

        op = tarantool.IPROTO_UPSERT
        sync = self.next_sync()
        schema_id = -1  # not sending schema_id with the request
        buf = WriteBuffer.new(self._encoding)
        buf.write_header(sync, op, schema_id)
        buf.encode_request_upsert(space, t, operations)
        buf.write_length()
        return self._protocol.execute(
            Request.new(op, sync, schema_id, space, push_subscribe),
            buf, timeout)

    cdef object _sql(self, str query, args, bint parse_metadata,
                      float timeout, bint push_subscribe):
        cdef:
            tarantool.iproto_type op
            uint64_t sync
            int64_t schema_id
            WriteBuffer buf
            Request req

        op = tarantool.IPROTO_EXECUTE
        sync = self.next_sync()
        schema_id = -1  # not sending schema_id with the request
        buf = WriteBuffer.new(self._encoding)
        buf.write_header(sync, op, schema_id)
        buf.encode_request_sql(query, args)
        buf.write_length()
        req = Request.new(op, sync, schema_id, None, push_subscribe)
        req.parse_metadata = parse_metadata
        req.parse_as_tuples = True
        return self._protocol.execute(req, buf, timeout)

    cdef object _auth(self, bytes salt, str username, str password,
                       float timeout, bint push_subscribe):
        cdef:
            tarantool.iproto_type op
            uint64_t sync
            int64_t schema_id
            WriteBuffer buf

            bytes username_bytes, password_bytes
            bytes hash1, hash2, scramble

        op = tarantool.IPROTO_AUTH
        sync = self.next_sync()
        schema_id = -1  # not sending schema_id with the request
        buf = WriteBuffer.new(self._encoding)
        buf.write_header(sync, op, schema_id)

        username_bytes = encode_unicode_string(username, self._encoding)
        password_bytes = encode_unicode_string(password, self._encoding)

        hash1 = Db._sha1((password_bytes,))
        hash2 = Db._sha1((hash1,))
        scramble = Db._sha1((salt, hash2))
        scramble = Db._strxor(hash1, scramble)

        buf.encode_request_auth(username_bytes, scramble)

        buf.write_length()
        return self._protocol.execute(
            Request.new(op, sync, schema_id, None, push_subscribe),
            buf, timeout)

    @staticmethod
    cdef bytes _sha1(tuple values):
        cdef object sha = hashlib.sha1()
        for i in values:
            if i is not None:
                sha.update(i)
        return sha.digest()

    @staticmethod
    cdef bytes _strxor(bytes hash1, bytes scramble):
        cdef:
            char *hash1_str
            ssize_t hash1_len

            char *scramble_str
            ssize_t scramble_len
        cpython.bytes.PyBytes_AsStringAndSize(hash1,
                                              &hash1_str, &hash1_len)
        cpython.bytes.PyBytes_AsStringAndSize(scramble,
                                              &scramble_str, &scramble_len)
        for i in range(scramble_len):
            scramble_str[i] = hash1_str[i] ^ scramble_str[i]
        return scramble

    # public methods

    def ping(self, timeout=-1):
        return self._ping(timeout, False)

    def call16(self, func_name, args=None, timeout=-1, push_subscribe=False):
        return self._call16(func_name, args, timeout, push_subscribe)

    def call(self, func_name, args=None, timeout=-1, push_subscribe=False):
        return self._call(func_name, args, timeout, push_subscribe)

    def eval(self, expression, args=None, timeout=-1, push_subscribe=False):
        return self._eval(expression, args, timeout, push_subscribe)

    def select(self, space, key=None,
               offset=0, limit=0xffffffff, index=0, iterator=0,
               timeout=-1):
        cdef:
            SchemaSpace sp
            SchemaIndex idx
        sp = self._protocol._schema.get_or_create_space(space)
        idx = sp.get_index(index)

        iterator = self._protocol.transform_iterator(iterator)
        if key is None and iterator == 0:
            iterator = 2  # ALL

        return self._select(sp, idx, key, offset, limit, iterator,
                            timeout, False)

    def insert(self, space, t, replace=False,
               timeout=-1):
        cdef:
            SchemaSpace sp
        sp = self._protocol._schema.get_or_create_space(space)

        return self._insert(sp, t, replace, timeout, False)

    def replace(self, space, t, timeout=-1):
        cdef:
            SchemaSpace sp
        sp = self._protocol._schema.get_or_create_space(space)

        return self._insert(sp, t, True, timeout, False)

    def delete(self, space, key, index=0, timeout=-1):
        cdef:
            SchemaSpace sp
            SchemaIndex idx
        sp = self._protocol._schema.get_or_create_space(space)
        idx = sp.get_index(index)

        return self._delete(sp, idx, key, timeout, False)

    def update(self, space, key, operations, index=0, timeout=-1):
        cdef:
            SchemaSpace sp
            SchemaIndex idx
        sp = self._protocol._schema.get_or_create_space(space)
        idx = sp.get_index(index)

        return self._update(sp, idx, key, operations, timeout, False)

    def upsert(self, space, t, operations, timeout=-1):
        cdef:
            SchemaSpace sp
        sp = self._protocol._schema.get_or_create_space(space)

        return self._upsert(sp, t, operations, timeout, False)

    def sql(self, query, args, parse_metadata=True, timeout=-1):
        return self._sql(query, args, parse_metadata, timeout, False)



from libc.stdint cimport uint64_t, uint32_t


cdef class Db:
    cdef:
        BaseProtocol _protocol
        bytes _encoding

    @staticmethod
    cdef inline Db new(BaseProtocol protocol)

    cdef inline uint64_t next_sync(self)

    cdef object _ping(self, float timeout, bint push_subscribe)

    cdef object _call16(self, str func_name, args,
                        float timeout, bint push_subscribe)

    cdef object _call(self, str func_name, args,
                      float timeout, bint push_subscribe)

    cdef object _eval(self, str expression, args,
                      float timeout, bint push_subscribe)

    cdef object _select(self, SchemaSpace space, SchemaIndex index, key,
                        uint64_t offset, uint64_t limit, uint32_t iterator,
                        float timeout, bint push_subscribe)

    cdef object _insert(self, SchemaSpace space, t, bint replace,
                        float timeout, bint push_subscribe)

    cdef object _delete(self, SchemaSpace space, SchemaIndex index, key,
                        float timeout, bint push_subscribe)

    cdef object _update(self, SchemaSpace space, SchemaIndex index, key,
                        list operations, float timeout, bint push_subscribe)

    cdef object _upsert(self, SchemaSpace space, t, list operations,
                        float timeout, bint push_subscribe)

    cdef object _sql(self, str query, args, bint parse_metadata,
                     float timeout, bint push_subscribe)

    cdef object _auth(self, bytes salt, str username, str password,
                      float timeout, bint push_subscribe)

    @staticmethod
    cdef bytes _sha1(tuple values)

    @staticmethod
    cdef bytes _strxor(bytes hash1, bytes scramble)



import asyncio
import sys
import warnings
from typing import Optional

PY_37 = sys.version_info >= (3, 7, 0)

if PY_37:
    __get_running_loop = asyncio.get_running_loop
else:
    def __get_running_loop() -> asyncio.AbstractEventLoop:
        loop = asyncio.get_event_loop()
        if not loop.is_running():
            raise RuntimeError('no running event loop')
        return loop


def get_running_loop(loop_arg: Optional[asyncio.AbstractEventLoop] = None):
    if loop_arg is not None:
        warnings.warn(
            'loop argument is deprecated and is dropped',
            DeprecationWarning,
            stacklevel=3
        )

    return __get_running_loop()



cdef enum iproto_header_key:
    IPROTO_REQUEST_TYPE = 0x00
    IPROTO_SYNC = 0x01
    IPROTO_REPLICA_ID = 0x02
    IPROTO_LSN = 0x03
    IPROTO_TIMESTAMP = 0x04
    IPROTO_SCHEMA_VERSION = 0x05
    IPROTO_SERVER_VERSION = 0x06
    IPROTO_GROUP_ID = 0x07


cdef enum iproto_key:
    IPROTO_SPACE_ID = 0x10
    IPROTO_INDEX_ID = 0x11
    IPROTO_LIMIT = 0x12
    IPROTO_OFFSET = 0x13
    IPROTO_ITERATOR = 0x14
    IPROTO_INDEX_BASE = 0x15

    IPROTO_KEY = 0x20
    IPROTO_TUPLE = 0x21
    IPROTO_FUNCTION_NAME = 0x22
    IPROTO_USER_NAME = 0x23
    IPROTO_INSTANCE_UUID = 0x24
    IPROTO_CLUSTER_UUID = 0x25
    IPROTO_VCLOCK = 0x26
    IPROTO_EXPR = 0x27
    IPROTO_OPS = 0x28
    IPROTO_BALLOT = 0x29
    IPROTO_TUPLE_META = 0x2a
    IPROTO_OPTIONS = 0x2b

    IPROTO_DATA = 0x30
    IPROTO_ERROR = 0x31
    IPROTO_METADATA = 0x32

    IPROTO_SQL_TEXT = 0x40
    IPROTO_SQL_BIND = 0x41
    IPROTO_SQL_INFO = 0x42


cdef enum iproto_metadata_key:
    IPROTO_FIELD_NAME = 0
    IPROTO_FIELD_TYPE = 1


cdef enum iproto_sql_info_key:
    SQL_INFO_ROW_COUNT = 0
    SQL_INFO_AUTOINCREMENT_IDS = 1


cdef enum iproto_type:
    IPROTO_OK = 0
    IPROTO_SELECT = 1
    IPROTO_INSERT = 2
    IPROTO_REPLACE = 3
    IPROTO_UPDATE = 4
    IPROTO_DELETE = 5
    IPROTO_CALL_16 = 6
    IPROTO_AUTH = 7
    IPROTO_EVAL = 8
    IPROTO_UPSERT = 9
    IPROTO_CALL = 10
    IPROTO_EXECUTE = 11
    IPROTO_PING = 64

    IPROTO_CHUNK = 128

cdef enum iproto_update_operation:
    IPROTO_OP_ADD = b'+'
    IPROTO_OP_SUB = b'-'
    IPROTO_OP_AND = b'&'
    IPROTO_OP_XOR = b'^'
    IPROTO_OP_OR = b'|'
    IPROTO_OP_DELETE = b'#'
    IPROTO_OP_INSERT = b'!'
    IPROTO_OP_ASSIGN = b'='
    IPROTO_OP_SPLICE = b':'



import copy


def get_complex_param(replace_bin=True, encoding='utf-8'):
    p = {
        'a': 1,
        'b': 2.5,
        'c': [1, 2, [4, 5], {'3': 17}],
        'd': {
            'k': 1,
            'l': (1, 2)
        },
        'e': b'1234567890',
        'f': [],
        'g': {},
        'h': 'some long text',
        'i': 'русский текст',
        'j': False,
        'k': True,
        'l': None
    }
    p_copy = copy.deepcopy(p)
    # tuples return as lists
    p_copy['d']['l'] = list(p_copy['d']['l'])

    if replace_bin:
        # For some reason Tarantool in call returns MP_STR instead of MP_BIN
        p_copy['e'] = p_copy['e'].decode(encoding)

    return p, p_copy


def get_big_param(size=1024, parts=3):
    assert parts > 0
    d = {}
    for p in range(parts):
        d['a' + str(p)] = 'x' * (size // parts)
    return d



# -*- coding: utf-8 -*-
import os
import re

from setuptools import Extension

from setuptools.command import build_ext as _build_ext

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

description = "A fast Tarantool Database connector for Python/asyncio."
try:
    import pypandoc
    long_description = pypandoc.convert_file('README.md', 'rst')
except (IOError, ImportError):
    long_description = description


def find_version():
    for line in open("asynctnt/__init__.py"):
        if line.startswith("__version__"):
            return re.match(
                r"""__version__\s*=\s*(['"])([^'"]+)\1""", line).group(2)


CYTHON_VERSION = '0.29.21'


class build_ext(_build_ext.build_ext):
    user_options = _build_ext.build_ext.user_options + [
        ('cython-always', None,
            'run cythonize() even if .c files are present'),
        ('cython-annotate', None,
            'Produce a colorized HTML version of the Cython source.'),
        ('cython-directives=', None,
            'Cython compiler directives'),
    ]

    def initialize_options(self):
        super(build_ext, self).initialize_options()
        self.cython_always = False
        self.cython_annotate = None
        self.cython_directives = None

    def finalize_options(self):
        need_cythonize = self.cython_always
        cfiles = {}

        for extension in self.distribution.ext_modules:
            for i, sfile in enumerate(extension.sources):
                if sfile.endswith('.pyx'):
                    prefix, ext = os.path.splitext(sfile)
                    cfile = prefix + '.c'

                    if os.path.exists(cfile) and not self.cython_always:
                        extension.sources[i] = cfile
                    else:
                        if os.path.exists(cfile):
                            cfiles[cfile] = os.path.getmtime(cfile)
                        else:
                            cfiles[cfile] = 0
                        need_cythonize = True

        if need_cythonize:
            try:
                import Cython
            except ImportError:
                raise RuntimeError(
                    'please install Cython to compile asynctnt from source')

            import pkg_resources
            cython_dep = pkg_resources.Requirement.parse(CYTHON_VERSION)
            if Cython.__version__ not in cython_dep:
                raise RuntimeError(
                    'asynctnt requires Cython version {}'.format(
                        CYTHON_VERSION))

            from Cython.Build import cythonize

            directives = {
                'language_level': '3'
            }
            if self.cython_directives:
                for directive in self.cython_directives.split(','):
                    k, _, v = directive.partition('=')
                    if v.lower() == 'false':
                        v = False
                    if v.lower() == 'true':
                        v = True

                    directives[k] = v

            self.distribution.ext_modules[:] = cythonize(
                self.distribution.ext_modules,
                compiler_directives=directives,
                annotate=self.cython_annotate)

        super(build_ext, self).finalize_options()


setup(
    name="asynctnt",
    packages=["asynctnt"],
    include_package_data=True,
    cmdclass={'build_ext': build_ext},
    ext_modules=[
        Extension("asynctnt.iproto.protocol",
                  sources=[
                      "asynctnt/iproto/protocol.pyx",
                      "third_party/msgpuck/msgpuck.c",
                      "third_party/msgpuck/hints.c",
                      "asynctnt/iproto/tupleobj/tupleobj.c"
                  ],
                  include_dirs=[
                      'third_party',
                      'asynctnt/iproto',
                  ])
    ],
    version=find_version(),
    author="igorcoding",
    author_email="igorcoding@gmail.com",
    url='https://github.com/igorcoding/asynctnt',
    license='Apache Software License',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        'Programming Language :: Python :: Implementation :: CPython',
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Database :: Front-Ends"
    ],
    install_requires=[
        "PyYAML >= 5.0"
    ],
    setup_requires=[
        "Cython=={}".format(CYTHON_VERSION)
    ],
    description=description,
    long_description=long_description,
    test_suite='run_tests.discover_tests'
)



import asyncio
import enum
import functools
import os
import typing
from typing import Optional

from .exceptions import TarantoolDatabaseError, \
    ErrorCode, TarantoolNotConnectedError
from .iproto import protocol
from .log import logger

__all__ = (
    'Connection', 'connect', 'ConnectionState'
)

from .utils import get_running_loop


class ConnectionState(enum.IntEnum):
    CONNECTING = 1
    CONNECTED = 2
    RECONNECTING = 3
    DISCONNECTING = 4
    DISCONNECTED = 5


_MethodRet = typing.Union[typing.Awaitable[protocol.Response], asyncio.Future]


class Connection:
    __slots__ = (
        '_host', '_port', '_username', '_password',
        '_fetch_schema', '_auto_refetch_schema', '_initial_read_buffer_size',
        '_encoding', '_connect_timeout', '_reconnect_timeout',
        '_request_timeout', '_ping_timeout', '_loop', '_state', '_state_prev',
        '_transport', '_protocol', '_db',
        '_disconnect_waiter', '_reconnect_task',
        '_connect_lock', '_disconnect_lock',
        '_ping_task', '__create_task'
    )

    def __init__(self, *,
                 host='127.0.0.1',
                 port=3301,
                 username=None,
                 password=None,
                 fetch_schema=True,
                 auto_refetch_schema=True,
                 connect_timeout=3.,
                 request_timeout=-1.,
                 reconnect_timeout=1. / 3.,
                 ping_timeout=5.,
                 encoding=None,
                 initial_read_buffer_size=None,
                 **kwargs):

        """
            Connection constructor.

            To manipulate a Connection instance there are several functions:

                * await connect() - performs connecting, authorization and
                                    schema fetching.

                * await disconnect() - performs disconnection.

                * close() - closes connection (not a coroutine)

            Connection also supports context manager protocol, which connects
            on entering and disconnecting on leaving a block.

            So one can simply use it as follows:

            .. code-block:: python

                async with asynctnt.Connection() as conn:
                    await conn.call('box.info')

            :param host:
                    Tarantool host (pass ``unix/`` to connect to unix socket)
            :param port:
                    Tarantool port
                    (pass ``/path/to/sockfile`` to connect ot unix socket)
            :param username:
                    Username to use for auth
                    (if ``None`` you are connected as a guest)
            :param password:
                    Password to use for auth
            :param fetch_schema:
                    Pass ``True`` to be able to use spaces and indexes names in
                    data manipulation routines (default is ``True``)
            :param auto_refetch_schema:
                    If set to ``True`` then when ER_WRONG_SCHEMA_VERSION error
                    occurs on a request, schema is refetched and the initial
                    request is resent. If set to ``False`` then schema will not
                    be checked by Tarantool, so no errors will occur
            :param connect_timeout:
                    Time in seconds how long to wait for connecting to socket
            :param request_timeout:
                    Request timeout (in seconds) for all requests
                    (by default there is no timeout)
            :param reconnect_timeout:
                    Time in seconds to wait before automatic reconnect
                    (set to ``0`` or ``None`` to disable auto reconnect)
            :param ping_timeout:
                    If specified (default is 5 seconds) a background task
                    will be created which will ping Tarantool instance
                    periodically to check if it is alive and update schema
                    if it is changed
                    (set to ``0`` or ``None`` to disable this task)
            :param encoding:
                    The encoding to use for all strings
                    encoding and decoding (default is ``utf-8``)
            :param initial_read_buffer_size:
                    Initial and minimum size of read buffer in bytes.
                    Higher value means less reallocations, but higher
                    memory usage (default is 131072).
            :param loop:
                    (Deprecated) Asyncio event loop to use
        """
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._fetch_schema = False if fetch_schema is None else fetch_schema
        if auto_refetch_schema:  # None hack
            self._auto_refetch_schema = True
            if not self._fetch_schema:
                logger.warning('Setting fetch_schema to True as '
                               'auto_refetch_schema is True')
                self._fetch_schema = True
        else:
            self._auto_refetch_schema = False
        self._initial_read_buffer_size = initial_read_buffer_size
        self._encoding = encoding or 'utf-8'

        self._connect_timeout = connect_timeout
        self._reconnect_timeout = reconnect_timeout or 0
        self._request_timeout = request_timeout
        self._ping_timeout = ping_timeout or 0

        self._loop = get_running_loop(kwargs.pop('loop', None))

        self._transport = None
        self._protocol = None
        self._db = _DbMock()

        self._state = ConnectionState.DISCONNECTED
        self._state_prev = ConnectionState.DISCONNECTED
        self._disconnect_waiter = None
        self._reconnect_task = None
        self._connect_lock = asyncio.Lock()
        self._disconnect_lock = asyncio.Lock()
        self._ping_task = None

        if hasattr(self._loop, 'create_task'):
            self.__create_task = self._loop.create_task
        else:
            self.__create_task = asyncio.ensure_future

    def _set_state(self, new_state):
        if self._state != new_state:
            logger.debug('Changing state %s -> %s',
                         self._state.name, new_state.name)
            self._state_prev = self._state
            self._state = new_state

    def connection_lost(self, exc):
        if self._transport:
            self._transport.close()
        self._transport = None

        if self._disconnect_waiter:
            # disconnect() call happened
            self._disconnect_waiter.set_result(True)
            return

        # connection lost

        if self._reconnect_timeout > 0:
            # should reconnect
            self._start_reconnect(return_exceptions=False)
        else:
            # should not reconnect, close everything
            self.close()

    async def _ping_task_func(self):
        while self._state == ConnectionState.CONNECTED:
            try:
                await self.ping(timeout=2.0)
            except asyncio.CancelledError:
                break
            except Exception:
                pass

            await asyncio.sleep(self._ping_timeout)

    def _start_reconnect(self, return_exceptions=False):
        if self._state in [ConnectionState.CONNECTING,
                           ConnectionState.RECONNECTING]:
            logger.debug('%s Cannot start reconnect: already reconnecting',
                         self.fingerprint)
            return

        if self._reconnect_task:
            return

        logger.info('%s Started reconnecting', self.fingerprint)
        self._set_state(ConnectionState.RECONNECTING)

        self._reconnect_task = self.__create_task(
            self._connect(return_exceptions=return_exceptions)
        )

    def protocol_factory(self, connected_fut, cls=protocol.Protocol):
        return cls(host=self._host,
                   port=self._port,
                   username=self._username,
                   password=self._password,
                   fetch_schema=self._fetch_schema,
                   auto_refetch_schema=self._auto_refetch_schema,
                   request_timeout=self._request_timeout,
                   initial_read_buffer_size=self._initial_read_buffer_size,
                   encoding=self._encoding,
                   connected_fut=connected_fut,
                   on_connection_made=None,
                   on_connection_lost=self.connection_lost,
                   loop=self._loop)

    async def _connect(self, return_exceptions=True):
        async with self._connect_lock:
            while True:
                try:
                    ignore_states = {
                        ConnectionState.CONNECTED,
                        ConnectionState.DISCONNECTING,  # disconnect() called
                    }

                    if self._state in ignore_states:
                        self._reconnect_task = None
                        return

                    self._set_state(ConnectionState.CONNECTING)

                    async def full_connect():
                        while True:
                            connected_fut = _create_future(self._loop)

                            if self._host.startswith('unix/'):
                                unix_path = self._port
                                assert isinstance(unix_path, str), \
                                    'port must be a str instance for ' \
                                    'unix socket'
                                assert unix_path, \
                                    'No unix file path specified'
                                assert os.path.exists(unix_path), \
                                    'Unix socket `{}` not found'.format(
                                        unix_path)

                                conn = self._loop.create_unix_connection(
                                    functools.partial(self.protocol_factory,
                                                      connected_fut),
                                    unix_path
                                )
                            else:
                                conn = self._loop.create_connection(
                                    functools.partial(self.protocol_factory,
                                                      connected_fut),
                                    self._host, self._port)

                            tr, pr = await conn

                            try:
                                timeout = 0.05  # wait at least something
                                if self._connect_timeout is not None:
                                    timeout = self._connect_timeout / 2

                                await asyncio.wait_for(connected_fut,
                                                       timeout=timeout)
                            except asyncio.TimeoutError:
                                tr.close()
                                continue  # try again
                            except Exception:
                                tr.close()
                                raise

                            return tr, pr

                    tr, pr = await asyncio.wait_for(
                        full_connect(),
                        timeout=self._connect_timeout,
                    )

                    logger.info('%s Connected successfully', self.fingerprint)
                    self._set_state(ConnectionState.CONNECTED)

                    self._transport = tr
                    self._protocol = pr
                    self._db = self._protocol.get_common_db()
                    self._reconnect_task = None
                    self._normalize_api()

                    if self._ping_timeout:
                        self._ping_task = \
                            self.__create_task(self._ping_task_func())
                    return
                except TarantoolDatabaseError as e:
                    skip_errors = {
                        ErrorCode.ER_LOADING,
                        ErrorCode.ER_NO_SUCH_SPACE,
                        ErrorCode.ER_NO_SUCH_INDEX
                    }
                    if e.code in skip_errors:
                        # If Tarantool is still loading then reconnect
                        if self._reconnect_timeout > 0:
                            await self._wait_reconnect(e)
                            continue

                    if return_exceptions:
                        self._reconnect_task = None
                        raise e

                    logger.exception(e)
                    if self._reconnect_timeout > 0:
                        await self._wait_reconnect(e)
                        continue

                    return  # no reconnect, no return_exceptions
                except asyncio.CancelledError:
                    logger.debug("connect is cancelled")
                    self._reconnect_task = None
                    raise
                except Exception as e:
                    if self._reconnect_timeout > 0:
                        await self._wait_reconnect(e)
                        continue

                    if return_exceptions:
                        self._reconnect_task = None
                        raise e

                    logger.exception(e)
                    return  # no reconnect, no return_exceptions

    async def _wait_reconnect(self, exc=None):
        self._set_state(ConnectionState.RECONNECTING)
        logger.warning('Connect to %s failed: %s. Retrying in %f seconds',
                       self.fingerprint,
                       repr(exc) if exc else '',
                       self._reconnect_timeout)

        await asyncio.sleep(self._reconnect_timeout)

    async def connect(self) -> 'Connection':
        """
            Connect coroutine
        """
        await self._connect(True)
        return self

    async def disconnect(self):
        """
            Disconnect coroutine
        """

        async with self._disconnect_lock:
            if self._state == ConnectionState.DISCONNECTED:
                return

            self._set_state(ConnectionState.DISCONNECTING)

            logger.info('%s Disconnecting...', self.fingerprint)
            if self._reconnect_task:
                self._reconnect_task.cancel()
                self._reconnect_task = None

            if self._ping_task and not self._ping_task.done():
                self._ping_task.cancel()
                self._ping_task = None

            self._db = _DbMock()
            if self._transport:
                self._disconnect_waiter = _create_future(self._loop)
                self._transport.close()
                self._transport = None
                self._protocol = None

                await self._disconnect_waiter
                self._disconnect_waiter = None
                self._set_state(ConnectionState.DISCONNECTED)
            else:
                self._transport = None
                self._protocol = None
                self._disconnect_waiter = None
                self._set_state(ConnectionState.DISCONNECTED)

    def close(self):
        """
            Same as disconnect, but not a coroutine, i.e. it does not wait
            for disconnect to finish.
        """

        if self._state == ConnectionState.DISCONNECTED:
            return

        self._set_state(ConnectionState.DISCONNECTING)
        logger.info('%s Disconnecting...', self.fingerprint)

        if self._reconnect_task and not self._reconnect_task.done():
            self._reconnect_task.cancel()
            self._reconnect_task = None

        if self._ping_task and not self._ping_task.done():
            self._ping_task.cancel()
            self._ping_task = None

        if self._transport:
            self._transport.close()

        self._transport = None
        self._protocol = None
        self._disconnect_waiter = None
        self._db = _DbMock()
        self._set_state(ConnectionState.DISCONNECTED)

    async def reconnect(self):
        """
            Reconnect coroutine.
            Just calls disconnect() and connect()
        """
        await self.disconnect()
        await self.connect()

    async def __aenter__(self):
        """
            Executed on entering the async with section.
            Connects to Tarantool instance.
        """
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
            Executed on leaving the async with section.
            Disconnects from Tarantool instance.
        """
        await self.disconnect()

    @property
    def fingerprint(self) -> str:
        return 'Tarantool[{}:{}]'.format(self._host, self._port)

    @property
    def host(self) -> str:
        """
            Tarantool host
        """
        return self._host

    @property
    def port(self) -> int:
        """
            Tarantool port
        """
        return self._port

    @property
    def username(self) -> Optional[str]:
        """
            Tarantool username
        """
        return self._username

    @property
    def password(self) -> Optional[str]:
        """
            Tarantool password
        """
        return self._password

    @property
    def fetch_schema(self) -> bool:
        """
            fetch_schema flag
        """
        return self._fetch_schema

    @property
    def auto_refetch_schema(self) -> bool:
        """
            auto_refetch_schema flag
        """
        return self._auto_refetch_schema

    @property
    def encoding(self) -> str:
        """
            Connection encoding
        """
        return self._encoding

    @property
    def reconnect_timeout(self) -> float:
        """
            Reconnect timeout value
        """
        return self._reconnect_timeout

    @property
    def connect_timeout(self) -> float:
        """
            Connect timeout value
        """
        return self._connect_timeout

    @property
    def request_timeout(self) -> float:
        """
            Request timeout value
        """
        return self._request_timeout

    @property
    def version(self) -> Optional[tuple]:
        """
            Protocol version tuple. ex.: (1, 6, 7)
        """
        if self._protocol is None:
            return None
        return self._protocol.get_version()

    @property
    def loop(self):
        """
            Asyncio event loop
        """
        return self._loop

    @property
    def state(self) -> ConnectionState:
        """
            Current connection state

            :rtype: ConnectionState
        """
        return self._state

    @property
    def is_connected(self) -> bool:
        """
            Check if an underlying connection is active
        """
        if self._protocol is None:
            return False
        return self._protocol.is_connected()

    @property
    def is_fully_connected(self) -> bool:
        """
            Check if connection is fully active (performed auth
            and schema fetching)
        """
        if self._protocol is None:
            return False
        return self._protocol.is_fully_connected()

    @property
    def schema_id(self) -> Optional[int]:
        """
            Tarantool's current schema id
        """
        if self._protocol is None:
            return None
        return self._protocol.schema_id

    @property
    def initial_read_buffer_size(self) -> int:
        """
            initial_read_buffer_size value
        """
        return self._initial_read_buffer_size

    async def refetch_schema(self):
        """
            Coroutine to force refetch schema
        """
        await self._protocol.refetch_schema()

    def ping(self, *, timeout=-1.0) -> _MethodRet:
        """
            Ping request coroutine

            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.ping(timeout=timeout)

    def call16(self, func_name, args=None, *,
               timeout=-1.0, push_subscribe=False) -> _MethodRet:
        """
            Call16 request coroutine. It is a call with an old behaviour
            (return result of a Tarantool procedure is wrapped into a tuple,
            if needed)

            :param func_name: function name to call
            :param args: arguments to pass to the function (list object)
            :param timeout: Request timeout
            :param push_subscribe: Subscribe to push notifications

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.call16(func_name, args,
                               timeout=timeout,
                               push_subscribe=push_subscribe)

    def call(self, func_name, args=None, *,
             timeout=-1.0, push_subscribe=False) -> _MethodRet:
        """
            Call request coroutine. It is a call with a new behaviour
            (return result of a Tarantool procedure is not wrapped into
            an extra tuple). If you're connecting to Tarantool with
            version < 1.7, then this call method acts like a call16 method

            Examples:

            .. code-block:: pycon

                # tarantool function:
                # function f(...)
                #     return ...
                # end

                >>> await conn.call('f')
                <Response sync=3 rowcount=0 data=[]>

                >>> await conn.call('f', [20, 42])
                <Response sync=3 rowcount=2 data=[20, 42]>

            :param func_name: function name to call
            :param args: arguments to pass to the function (list object)
            :param timeout: Request timeout
            :param push_subscribe: Subscribe to push notifications

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.call(func_name, args,
                             timeout=timeout, push_subscribe=push_subscribe)

    def eval(self, expression, args=None, *,
             timeout=-1.0, push_subscribe=False) -> _MethodRet:
        """
            Eval request coroutine.

            Examples:

            .. code-block:: pycon

                >>> await conn.eval('return 42')
                <Response sync=3 rowcount=1 data=[42]>


                >>> await conn.eval('return box.info.version')
                <Response sync=3 rowcount=1 data=['2.1.1-7-gd381a45b6']>

            :param expression: expression to execute
            :param args: arguments to pass to the function, that will
                         execute your expression (list object)
            :param timeout: Request timeout
            :param push_subscribe: Subscribe to push messages

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.eval(expression, args,
                             timeout=timeout, push_subscribe=push_subscribe)

    def select(self, space, key=None, **kwargs) -> _MethodRet:
        """
            Select request coroutine.

            Examples:

            .. code-block:: pycon

                >>> await conn.select('tester')
                <Response sync=3 rowcount=2 data=[
                    <TarantoolTuple id=1 name='one'>,
                    <TarantoolTuple id=2 name='two'>
                ]>

                >>> res = await conn.select('_space', ['tester'], index='name')
                >>> res.data
                [<TarantoolTuple id=512
                                 owner=1
                                 name='tester'
                                 engine='memtx'
                                 field_count=0
                                 flags={}
                                 format=[
                                    {'name': 'id', 'type': 'unsigned'},
                                    {'name': 'name', 'type': 'string'}
                                 ]>]


            :param space: space id or space name.
            :param key: key to select
            :param offset: offset to use
            :param limit: limit to use
            :param index: index id or name
            :param iterator: one of the following

                        * iterator id (int number),
                        * :class:`asynctnt.Iterator` object
                        * string with an iterator name

            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.select(space, key, **kwargs)

    def insert(self, space, t, *, replace=False, timeout=-1) -> _MethodRet:
        """
            Insert request coroutine.

            Examples:

            .. code-block:: pycon

                # Basic usage
                >>> await conn.insert('tester', [0, 'hello'])
                <Response sync=3 rowcount=1 data=[
                    <TarantoolTuple id=0 name='hello'>
                ]>

                # Using dict as an argument tuple
                >>> await conn.insert('tester', {
                ...                     'id': 0
                ...                     'text': 'hell0'
                ...                   })
                <Response sync=3 rowcount=1 data=[
                    <TarantoolTuple id=0 name='hello'>
                ]>

            :param space: space id or space name.
            :param t: tuple to insert (list object)
            :param replace: performs replace request instead of insert
            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.insert(space, t,
                               replace=replace,
                               timeout=timeout)

    def replace(self, space, t, *, timeout=-1.0) -> _MethodRet:
        """
            Replace request coroutine. Same as insert, but replace.

            :param space: space id or space name.
            :param t: tuple to insert (list object)
            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.replace(space, t, timeout=timeout)

    def delete(self, space, key, **kwargs) -> _MethodRet:
        """
            Delete request coroutine.

            Examples:

            .. code-block:: pycon

                # Assuming tuple [0, 'hello'] is in space tester

                >>> await conn.delete('tester', [0])
                <Response sync=3 rowcount=1 data=[
                    <TarantoolTuple id=0 name='hello'>
                ]>

            :param space: space id or space name.
            :param key: key to delete
            :param index: index id or name
            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.delete(space, key, **kwargs)

    def update(self, space, key, operations, **kwargs) -> _MethodRet:
        """
            Update request coroutine.

            Examples:

            .. code-block:: pycon

                # Assuming tuple [0, 'hello'] is in space tester

                >>> await conn.update('tester', [0], [ ['=', 1, 'hi!'] ])
                <Response sync=3 rowcount=1 data=[
                    <TarantoolTuple id=0 name='hi!'>
                ]>

                # you can use fields names as well
                >>> res = await conn.update('tester', [0],
                ...                         [ ['=', 'text', 'hola'] ])
                <Response sync=3 rowcount=1 data=[
                    <TarantoolTuple id=0 name='hola'>
                ]>

            :param space: space id or space name.
            :param key: key to update
            :param operations:
                    Operations list of the following format:
                    [ [op_type, field_no, ...], ... ]. Please refer to
                    https://tarantool.org/doc/book/box/box_space.html?highlight=update#lua-function.space_object.update
                    You can use field numbers as well as their names in space
                    format as a field_no (if only fetch_schema is True).
                    If field is unknown then TarantoolSchemaError is raised.
            :param index: index id or name
            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.update(space, key, operations, **kwargs)

    def upsert(self, space, t, operations, **kwargs) -> _MethodRet:
        """
            Update request coroutine. Performs either insert or update
            (depending of either tuple exists or not)

            Examples:

            .. code-block:: pycon

                # upsert does not return anything
                >>> await conn.upsert('tester', [0, 'hello'],
                ...                   [ ['=', 1, 'hi!'] ])
                <Response sync=3 rowcount=0 data=[]>

            :param space: space id or space name.
            :param t: tuple to insert if it's not in space
            :param operations:
                    Operations list to use for update if tuple is already in
                    space. It has the same format as in update requets:
                    [ [op_type, field_no, ...], ... ]. Please refer to
                    https://tarantool.org/doc/book/box/box_space.html?highlight=update#lua-function.space_object.update
                    You can use field numbers as well as their names in space
                    format as a field_no (if only fetch_schema is True).
                    If field is unknown then TarantoolSchemaError is raised.
            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.upsert(space, t, operations, **kwargs)

    def sql(self, query, args=None, *,
            parse_metadata=True, timeout=-1.0) -> _MethodRet:
        """
            Executes an SQL statement (only for Tarantool > 2)

            Examples:

            .. code-block:: pycon

                >>> await conn.sql("select 1 as a, 2 as b")
                <Response sync=3 rowcount=1 data=[<TarantoolTuple A=1 B=2>]>

                >>> await conn.sql("select * from sql_space")
                <Response sync=3 rowcount=2 data=[
                    <TarantoolTuple ID=1 NAME='James Bond'>,
                    <TarantoolTuple ID=2 NAME='Ethan Hunt'>
                ]>

                >>> await conn.sql("select * from sql_space",
                ...                parse_metadata=False)
                <Response sync=3 rowcount=2 data=[
                    <TarantoolTuple 0=1 1='James Bond'>,
                    <TarantoolTuple 0=2 1='Ethan Hunt'>
                ]>

            :param query: SQL query
            :param args: Query arguments
            :param parse_metadata: Set to False to disable response's metadata
                                   parsing for better performance
            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.sql(query, args,
                            parse_metadata=parse_metadata,
                            timeout=timeout)

    def _normalize_api(self):
        if (1, 6) <= self.version < (1, 7):  # pragma: nocover
            Connection.call = Connection.call16

    def __repr__(self):
        return "<asynctnt.Connection host={} port={} state={}>".format(
            self.host,
            self.port,
            repr(self.state)
        )


def _create_future(loop):
    try:
        return loop.create_future()
    except AttributeError:  # pragma: nocover
        return asyncio.Future(loop=loop)


class _DbMock:
    def __getattr__(self, item):
        raise TarantoolNotConnectedError('Tarantool is not connected')


async def connect(**kwargs) -> Connection:
    """
        connect shorthand. See :class:`asynctnt.Connection` for kwargs details

        :return: :class:`asynctnt.Connection` object
    """
    return await Connection(**kwargs).connect()
